from openmdao.api import Group, IndepVarComp, ExplicitComponent
import numpy as np

from wisdem.commonse.utilities import nodal2sectional, sectional2nodal, assembleI, unassembleI, sectionalInterp
import wisdem.commonse.frustum as frustum
import wisdem.commonse.manufacturing as manufacture
from wisdem.commonse.UtilizationSupplement import shellBuckling_withStiffeners, GeometricConstraints
from wisdem.commonse import gravity, eps, AeroHydroLoads, CylinderWindDrag, CylinderWaveDrag
from wisdem.commonse.vertical_cylinder import CylinderDiscretization, CylinderMass
from wisdem.commonse.environment import PowerWind, LinearWaves

def get_inner_radius(Ro, t):
    # Radius varies at nodes, t varies by section
    return (Ro-sectional2nodal(t))

def I_tube(r_i, r_o, h, m):
    if type(r_i) == type(np.array([])):
        n = r_i.size
        r_i = r_i.flatten()
        r_o = r_o.flatten()
        h   = h.flatten()
        m   = m.flatten()
    else:
        n = 1
    Ixx = Iyy = (m/12.0) * (3.0*(r_i**2.0 + r_o**2.0) + h**2.0)
    Izz = 0.5 * m * (r_i**2.0 + r_o**2.0)
    return np.c_[Ixx, Iyy, Izz, np.zeros((n,3))]
    

class BulkheadProperties(ExplicitComponent):

    def initialize(self):
        self.options.declare('nSection')
        self.options.declare('nFull')
        
    def setup(self):
        nSection = self.options['nSection']
        nFull    = self.options['nFull']
        
        self.bulk_full = np.zeros( nFull, dtype=np.int_)

        self.add_input('z_full', val=np.zeros(nFull), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('z_param', val=np.zeros((nSection+1,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('d_full', val=np.zeros(nFull), units='m', desc='cylinder diameter at corresponding locations')
        self.add_input('t_full', val=np.zeros(nFull-1), units='m', desc='shell thickness at corresponding locations')
        self.add_input('rho', val=0.0, units='kg/m**3', desc='material density')
        self.add_input('bulkhead_thickness', val=np.zeros(nSection+1), units='m', desc='Nodal locations of bulkhead thickness, zero meaning no bulkhead, bottom to top (length = nsection + 1)')
        self.add_input('bulkhead_mass_factor', val=0.0, desc='Bulkhead mass correction factor')
        
        self.add_input('shell_mass', val=np.zeros(nFull-1), units='kg', desc='mass of column shell')
        self.add_input('material_cost_rate', 0.0, units='USD/kg', desc='Raw material cost rate: steel $1.1/kg, aluminum $3.5/kg')
        self.add_input('labor_cost_rate', 0.0, units='USD/min', desc='Labor cost rate')
        self.add_input('painting_cost_rate', 0.0, units='USD/m/m', desc='Painting / surface finishing cost rate')

        self.add_output('bulkhead_mass', val=np.zeros(nFull), units='kg', desc='mass of column bulkheads')
        self.add_output('bulkhead_cost', val=0.0, units='USD', desc='cost of column bulkheads')
        self.add_output('bulkhead_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of bulkheads relative to keel point')
        
        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs):
        # Unpack variables
        z_full     = inputs['z_full'] # at section nodes
        z_param    = inputs['z_param']
        R_od       = 0.5*inputs['d_full'] # at section nodes
        twall      = inputs['t_full'] # at section nodes
        R_id       = get_inner_radius(R_od, twall)
        t_bulk     = inputs['bulkhead_thickness'] # at section nodes
        rho        = inputs['rho']
        
        # Map bulkhead locations to finer computation grid
        Zf,Zp = np.meshgrid(z_full, z_param)
        idx = np.argmin( np.abs(Zf-Zp), axis=1 )
        t_bulk_full = np.zeros( z_full.shape )
        t_bulk_full[idx] = t_bulk
        # Make sure top and bottom are capped
        if (t_bulk_full[ 0] == 0.0): t_bulk_full[ 0] = twall[ 0]
        if (t_bulk_full[-1] == 0.0): t_bulk_full[-1] = twall[-1]
        
        # Compute bulkhead volume at every section node
        # Assume bulkheads are same thickness as shell wall
        V_bulk = np.pi * R_id**2 * t_bulk_full

        # Convert to mass with fudge factor for design features not captured in this simple approach
        m_bulk = inputs['bulkhead_mass_factor'] * rho * V_bulk

        # Compute moments of inertia at keel
        # Assume bulkheads are just simple thin discs with radius R_od-t_wall and mass already computed
        Izz = 0.5 * m_bulk * R_id**2
        Ixx = Iyy = 0.5 * Izz
        I_keel = np.zeros((3,3))
        dz  = z_full - z_full[0]
        for k in range(m_bulk.size):
            R = np.array([0.0, 0.0, dz[k]])
            Icg = assembleI( [Ixx[k], Iyy[k], Izz[k], 0.0, 0.0, 0.0] )
            I_keel += Icg + m_bulk[k]*(np.dot(R, R)*np.eye(3) - np.outer(R, R))


        # Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
        # All dimensions for correlations based on mm, not meters.
        k_m     = inputs['material_cost_rate'] #1.1 # USD / kg carbon steel plate
        k_f     = inputs['labor_cost_rate'] #1.0 # USD / min labor
        k_p     = inputs['painting_cost_rate'] #USD / m^2 painting
        m_shell = inputs['shell_mass'].sum()
        nbulk   = np.count_nonzero(V_bulk)
        bulkind = (V_bulk > 0.0)
        
        # Cost Step 1) Cutting flat plates using plasma cutter
        cutLengths = 2.0 * np.pi * R_id * bulkind
        # Cost Step 2) Fillet welds with GMAW-C (gas metal arc welding with CO2) of bulkheads to shell
        theta_w = 3.0 # Difficulty factor

        # Labor-based expenses
        K_f = k_f * ( manufacture.steel_cutting_plasma_time(cutLengths, t_bulk_full) +
                      manufacture.steel_filett_welding_time(theta_w, nbulk, m_bulk+m_shell, 2*np.pi*R_id, t_bulk_full) )
        
        # Cost Step 3) Painting (two sided)
        theta_p = 1.0
        K_p  = k_p * theta_p * 2 * (np.pi * R_id**2.0 * bulkind).sum()

        # Material cost, without outfitting
        K_m = k_m * m_bulk.sum()

        # Total cost
        c_bulk = K_m + K_f + K_p
        
        # Store results
        outputs['bulkhead_I_keel'] = unassembleI(I_keel)
        outputs['bulkhead_mass'] = m_bulk
        outputs['bulkhead_cost'] = c_bulk

    

class BuoyancyTankProperties(ExplicitComponent):

    def initialize(self):
        self.options.declare('nFull')
        
    def setup(self):
        nFull    = self.options['nFull']
        
        self.add_input('d_full', val=np.zeros(nFull), units='m', desc='cylinder diameter at corresponding locations')
        self.add_input('z_full', val=np.zeros(nFull), units='m', desc='z-coordinates of section nodes')
        self.add_input('rho', val=0.0, units='kg/m**3', desc='material density')

        self.add_input('shell_mass', val=np.zeros(nFull-1), units='kg', desc='mass of column shell')
        self.add_input('material_cost_rate', 0.0, units='USD/kg', desc='Raw material cost: steel $1.1/kg, aluminum $3.5/kg')
        self.add_input('labor_cost_rate', 0.0, units='USD/min', desc='Labor cost')
        self.add_input('painting_cost_rate', 0.0, units='USD/m/m', desc='Painting / surface finishing cost rate')
        
        self.add_input('buoyancy_tank_diameter', val=0.0, units='m', desc='Radius of heave plate at bottom of column')
        self.add_input('buoyancy_tank_height', val=0.0, units='m', desc='Radius of heave plate at bottom of column')
        self.add_input('buoyancy_tank_location', val=0.0, units='m', desc='Radius of heave plate at bottom of column')
        self.add_input('buoyancy_tank_mass_factor', val=0.0, desc='Heave plate mass correction factor')

        self.add_output('buoyancy_tank_mass', val=0.0, units='kg', desc='mass of buoyancy tank')
        self.add_output('buoyancy_tank_cost', val=0.0, units='USD', desc='cost of buoyancy tank')
        self.add_output('buoyancy_tank_cg', val=0.0, units='m', desc='z-coordinate of center of mass for buoyancy tank')
        self.add_output('buoyancy_tank_displacement', val=0.0, units='m**3', desc='volume of water displaced by buoyancy tank')
        self.add_output('buoyancy_tank_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of heave plate relative to keel point')

        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs):
        # Unpack variables
        z_full    = inputs['z_full']

        R_od      = 0.5*inputs['d_full']
        R_plate   = 0.5*inputs['buoyancy_tank_diameter']
        h_box     = inputs['buoyancy_tank_height']

        location  = inputs['buoyancy_tank_location']
        
        coeff     = inputs['buoyancy_tank_mass_factor']
        rho       = inputs['rho']

        # Current hard-coded, coarse specification of shell thickness
        t_plate   = R_plate / 50.0

        # Z-locations of buoyancy tank
        z_lower   = location * (z_full[-1] - z_full[0]) + z_full[0]
        z_cg      = z_lower + 0.5*h_box
        z_upper   = z_lower +     h_box

        # Mass and volume properties that subtract out central column contributions for no double-counting
        R_col     = np.interp([z_lower, z_upper], z_full, R_od)
        if not np.any(R_plate > R_col): R_plate = 0.0
        A_plate   = np.maximum(0.0, np.pi * (R_plate**2.0 - R_col**2.0))
        m_plate   = coeff * rho * t_plate * A_plate
        A_box     = A_plate.sum() + 2.0 * np.pi * R_plate * h_box
        m_box     = coeff * rho * t_plate * A_box

        # Compute displcement for buoyancy calculations, but check for what is submerged
        V_box      = np.pi * R_plate**2.0 * h_box
        V_box     -= frustum.frustumVol(R_col[0], R_col[1], h_box)
        if z_lower >= 0.0:
            V_box  = 0.0
        elif z_upper >= 0.0:
            V_box *= (- z_lower / h_box)
        V_box      = np.maximum(0.0, V_box)

        # Now do moments of inertia
        # First find MoI at cg of all components
        R_plate += eps
        Ixx_box   = frustum.frustumShellIxx(R_plate, R_plate, t_plate, h_box)
        Izz_box   = frustum.frustumShellIzz(R_plate, R_plate, t_plate, h_box)
        I_plateL  = 0.25 * m_plate[0] * (R_plate**2.0 - R_col[0]**2.0) * np.array([1.0, 1.0, 2.0, 0.0, 0.0, 0.0])
        I_plateU  = 0.25 * m_plate[1] * (R_plate**2.0 - R_col[1]**2.0) * np.array([1.0, 1.0, 2.0, 0.0, 0.0, 0.0])

        # Move to keel for consistency
        I_keel    = np.zeros((3,3))
        if R_plate > eps:
            # Add in lower plate
            r         = np.array([0.0, 0.0, z_lower])
            Icg       = assembleI( I_plateL )
            I_keel   += Icg + m_plate[0]*(np.dot(r, r)*np.eye(3) - np.outer(r, r))
            # Add in upper plate
            r         = np.array([0.0, 0.0, z_upper])
            Icg       = assembleI( I_plateU )
            I_keel   += Icg + m_plate[1]*(np.dot(r, r)*np.eye(3) - np.outer(r, r))
            # Add in box cylinder
            r         = np.array([0.0, 0.0, z_cg])
            Icg       = assembleI( [Ixx_box, Ixx_box, Izz_box, 0.0, 0.0, 0.0] )
            I_keel   += Icg + m_plate[1]*(np.dot(r, r)*np.eye(3) - np.outer(r, r))


        # Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
        # All dimensions for correlations based on mm, not meters.
        k_m     = inputs['material_cost_rate'] #1.1 # USD / kg carbon steel plate
        k_f     = inputs['labor_cost_rate'] #1.0 # USD / min labor
        k_p     = inputs['painting_cost_rate'] #USD / m^2 painting
        m_shell = inputs['shell_mass'].sum()

        # Cost Step 1) Cutting flat plates using plasma cutter into box plate sizes
        cutLengths = 2.0 * np.pi * (3.0*R_plate + R_col.sum()) # x3 for two plates + side wall
        # Cost Step 2) Welding box plates together GMAW-C (gas metal arc welding with CO2) fillet welds
        theta_w = 3.0 # Difficulty factor
        # Cost Step 3) Welding box to shell GMAW-C (gas metal arc welding with CO2) fillet welds

        # Labor-based expenses
        K_f = k_f * ( manufacture.steel_cutting_plasma_time(cutLengths, t_plate) +
                      manufacture.steel_filett_welding_time(theta_w, 3.0, m_box, 2*np.pi*R_plate, t_plate) +
                      manufacture.steel_filett_welding_time(theta_w, 2.0, m_box+m_shell, 2*np.pi*R_col, t_plate) )
        
        # Cost Step 4) Painting
        theta_p = 1.5
        K_p  = k_p * theta_p * 2.0 * A_box
        
        # Material cost, without outfitting
        K_m = k_m * m_box

        # Total cost
        c_box = K_m + K_f + K_p
            
        # Store outputs
        outputs['buoyancy_tank_cost']         = c_box
        outputs['buoyancy_tank_mass']         = m_box
        outputs['buoyancy_tank_cg']           = z_cg
        outputs['buoyancy_tank_displacement'] = V_box
        outputs['buoyancy_tank_I_keel']       = unassembleI(I_keel)
        
        
class StiffenerProperties(ExplicitComponent):
    """Computes column stiffener properties by section.  
    Stiffener being the ring of T-cross section members placed periodically along column
    Assumes constant stiffener spacing along the column, but allows for varying stiffener geometry
    Slicing the column lengthwise would reveal the stiffener T-geometry as:
    |              |
    |              |  
    |   |      |   |
    |----      ----|
    |   |      |   |
    |              |
    |              |
    """

    def initialize(self):
        self.options.declare('nSection')
        self.options.declare('nFull')
        
    def setup(self):
        nSection = self.options['nSection']
        nFull    = self.options['nFull']

        self.add_input('d_full', val=np.zeros(nFull), units='m', desc='cylinder diameter at corresponding locations')
        self.add_input('t_full', val=np.zeros(nFull-1), units='m', desc='shell thickness at corresponding locations')
        self.add_input('z_full', val=np.zeros(nFull), units='m', desc='z-coordinates of section nodes')
        self.add_input('rho', val=0.0, units='kg/m**3', desc='material density')
        
        self.add_input('shell_mass', val=np.zeros(nFull-1), units='kg', desc='mass of column shell')
        self.add_input('material_cost_rate', 0.0, units='USD/kg', desc='Raw material cost: steel $1.1/kg, aluminum $3.5/kg')
        self.add_input('labor_cost_rate', 0.0, units='USD/min', desc='Labor cost')
        self.add_input('painting_cost_rate', 0.0, units='USD/m/m', desc='Painting / surface finishing cost rate')

        self.add_input('h_web', val=np.zeros((nFull-1,)), units='m', desc='height of stiffener web (base of T) within each section bottom to top')
        self.add_input('t_web', val=np.zeros((nFull-1,)), units='m', desc='thickness of stiffener web (base of T) within each section bottom to top')
        self.add_input('w_flange', val=np.zeros((nFull-1,)), units='m', desc='height of stiffener flange (top of T) within each section bottom to top')
        self.add_input('t_flange', val=np.zeros((nFull-1,)), units='m', desc='thickness of stiffener flange (top of T) within each section bottom to top')
        self.add_input('L_stiffener', val=np.zeros((nFull-1,)), units='m', desc='Axial distance from one ring stiffener to another within each section bottom to top')

        self.add_input('ring_mass_factor', val=0.0, desc='Stiffener ring mass correction factor')
        
        self.add_output('stiffener_mass', val=np.zeros(nFull-1), units='kg', desc='mass of column stiffeners')
        self.add_output('stiffener_cost', val=0.0, units='USD', desc='cost of column stiffeners')
        self.add_output('stiffener_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of stiffeners relative to keel point')
        self.add_output('number_of_stiffeners', val=np.zeros(nSection, dtype=np.int_), desc='number of stiffeners in each section')
        self.add_output('flange_spacing_ratio', val=np.zeros((nFull-1,)), desc='ratio between flange and stiffener spacing')
        self.add_output('stiffener_radius_ratio', val=np.zeros((nFull-1,)), desc='ratio between stiffener height and radius')

        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs):
        # Unpack variables
        R_od         = 0.5*inputs['d_full']
        t_wall       = inputs['t_full']
        z_full       = inputs['z_full'] # at section nodes
        h_section    = np.diff(z_full)
        V_shell      = frustum.frustumShellVol(R_od[:-1], R_od[1:], t_wall, h_section)
        R_od,_       = nodal2sectional( R_od ) # at section nodes
        
        t_web        = inputs['t_web']
        t_flange     = inputs['t_flange']
        h_web        = inputs['h_web']
        w_flange     = inputs['w_flange']
        L_stiffener  = inputs['L_stiffener']

        rho          = inputs['rho']
        
        # Outer and inner radius of web by section
        R_wo = R_od - t_wall
        R_wi = R_wo - h_web
        # Outer and inner radius of flange by section
        R_fo = R_wi
        R_fi = R_fo - t_flange

        # Material volumes by section
        V_web    = np.pi*(R_wo**2 - R_wi**2) * t_web
        V_flange = np.pi*(R_fo**2 - R_fi**2) * w_flange

        # Ring mass by volume by section 
        # Include fudge factor for design features not captured in this simple approach
        m_web    = inputs['ring_mass_factor'] * rho * V_web
        m_flange = inputs['ring_mass_factor'] * rho * V_flange
        m_ring   = m_web + m_flange
        n_stiff  = np.zeros(h_web.shape, dtype=np.int_)
        
        # Compute moments of inertia for stiffeners (lumped by section for simplicity) at keel
        I_web     = I_tube(R_wi, R_wo, t_web   , m_web)
        I_flange  = I_tube(R_fi, R_fo, w_flange, m_flange)
        I_ring    = I_web + I_flange
        I_keel    = np.zeros((3,3))

        # Now march up the column, adding stiffeners at correct spacing until we are done
        z_stiff  = []
        isection = 0
        epsilon  = 1e-6
        while True:
            if len(z_stiff) == 0:
                z_march = np.minimum(z_full[isection+1], z_full[0] + 0.5*L_stiffener[isection]) + epsilon
            else:
                z_march = np.minimum(z_full[isection+1], z_stiff[-1] + L_stiffener[isection]) + epsilon
            if z_march >= z_full[-1]: break
            
            isection = np.searchsorted(z_full, z_march) - 1
            
            if len(z_stiff) == 0:
                add_stiff = (z_march - z_full[0]) >= 0.5*L_stiffener[isection]
            else:
                add_stiff = (z_march - z_stiff[-1]) >= L_stiffener[isection]
                
            if add_stiff:
                z_stiff.append(z_march)
                n_stiff[isection] += 1
                
                R       = np.array([0.0, 0.0, (z_march - z_full[0])])
                Icg     = assembleI( I_ring[isection,:] )
                I_keel += Icg + m_ring[isection]*(np.dot(R, R)*np.eye(3) - np.outer(R, R))

        # Number of stiffener rings per section (height of section divided by spacing)
        outputs['stiffener_mass'] =  n_stiff * m_ring

        # Find total number of stiffeners in each original section
        nSection    = self.options['nSection']
        npts_per    = int(h_web.size / nSection)
        n_stiff_sec = np.zeros(nSection)
        for k in range(npts_per):
            n_stiff_sec += n_stiff[k::npts_per]
        outputs['number_of_stiffeners'] = n_stiff_sec


        # Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
        # All dimensions for correlations based on mm, not meters.
        k_m     = inputs['material_cost_rate'] #1.1 # USD / kg carbon steel plate
        k_f     = inputs['labor_cost_rate'] #1.0 # USD / min labor
        k_p     = inputs['painting_cost_rate'] #USD / m^2 painting
        m_shell = inputs['shell_mass'].sum()
        
        # Cost Step 1) Cutting stiffener strips from flat plates using plasma cutter
        cutLengths_w = 2.0 * np.pi * 0.5 * (R_wo + R_wi)
        cutLengths_f = 2.0 * np.pi * R_fo
        # Cost Step 2) Welding T-stiffeners together GMAW-C (gas metal arc welding with CO2) fillet welds
        theta_w = 3.0 # Difficulty factor
        # Cost Step 3) Welding stiffeners to shell GMAW-C (gas metal arc welding with CO2) fillet welds
        # Will likely fillet weld twice (top & bottom), so factor of 2 on second welding terms

        # Labor-based expenses
        K_f = k_f * ( manufacture.steel_cutting_plasma_time(n_stiff * cutLengths_w, t_web) +
                      manufacture.steel_cutting_plasma_time(n_stiff * cutLengths_f, t_flange) +
                      manufacture.steel_filett_welding_time(theta_w, n_stiff, m_ring, 2*np.pi*R_fo, t_web) +
                      manufacture.steel_filett_welding_time(theta_w, n_stiff, m_ring+m_shell, 2*np.pi*R_wo, t_web) )
        
        # Cost Step 4) Painting
        theta_p = 2.0
        K_p  = k_p * theta_p * (n_stiff*(2*np.pi*(R_wo**2.0-R_wi**2.0) + 2*np.pi*0.5*(R_fo+R_fi)*(2*w_flange + 2*t_flange) - 2*np.pi*R_fo*t_web)).sum()
        
        # Material cost, without outfitting
        K_m = k_m * outputs['stiffener_mass'].sum()

        # Total cost
        c_ring = K_m + K_f + K_p
        
        # Store results
        outputs['stiffener_cost'] = c_ring
        outputs['stiffener_I_keel'] = unassembleI(I_keel)
        
        # Create some constraints for reasonable stiffener designs for an optimizer
        outputs['flange_spacing_ratio']   = w_flange / (0.5*L_stiffener)
        outputs['stiffener_radius_ratio'] = (h_web + t_flange + t_wall) / R_od



        
class BallastProperties(ExplicitComponent):

    def initialize(self):
        self.options.declare('nFull')
        
    def setup(self):
        nFull    = self.options['nFull']

        self.add_input('water_density', val=0.0, units='kg/m**3', desc='density of water')
        self.add_input('d_full', val=np.zeros(nFull), units='m', desc='cylinder diameter at corresponding locations')
        self.add_input('t_full', val=np.zeros(nFull-1), units='m', desc='shell thickness at corresponding locations')
        self.add_input('z_full', val=np.zeros(nFull), units='m', desc='z-coordinates of section nodes')
        self.add_input('permanent_ballast_density', val=0.0, units='kg/m**3', desc='density of permanent ballast')
        self.add_input('permanent_ballast_height', val=0.0, units='m', desc='height of permanent ballast')
        self.add_input('ballast_cost_rate', val=0.0, units='USD/kg', desc='Cost per unit mass of ballast')

        self.add_output('ballast_cost', val=0.0, units='USD', desc='cost of permanent ballast')
        self.add_output('ballast_mass', val=np.zeros(nFull-1), units='kg', desc='mass of permanent ballast')
        self.add_output('ballast_z_cg', val=0.0, units='m', desc='z-coordinate or permanent ballast center of gravity')
        self.add_output('ballast_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of permanent ballast relative to keel point')
        self.add_output('variable_ballast_interp_zpts', val=np.zeros((nFull,)), units='m', desc='z-points of potential ballast mass')
        self.add_output('variable_ballast_interp_radius', val=np.zeros((nFull,)), units='m', desc='inner radius of column at potential ballast mass')

        
    def compute(self, inputs, outputs):
        # Unpack variables
        R_od        = 0.5*inputs['d_full']
        t_wall      = inputs['t_full']
        z_nodes     = inputs['z_full']
        h_ballast   = float(inputs['permanent_ballast_height'])
        rho_ballast = float(inputs['permanent_ballast_density'])
        rho_water   = float(inputs['water_density'])
        R_id_orig   = get_inner_radius(R_od, t_wall)

        npts = R_od.size
        section_mass = np.zeros(npts-1)
        
        # Geometry of the column in our coordinate system (z=0 at waterline)
        z_draft   = z_nodes[0]

        # Fixed and total ballast mass and cg
        # Assume they are bottled in columns a the keel of the column- first the permanent then the fixed
        zpts      = np.linspace(z_draft, z_draft+h_ballast, npts)
        R_id      = np.interp(zpts, z_nodes, R_id_orig)
        V_perm    = np.pi * np.trapz(R_id**2, zpts)
        m_perm    = rho_ballast * V_perm
        z_cg_perm = rho_ballast * np.pi * np.trapz(zpts*R_id**2, zpts) / m_perm if m_perm > 0.0 else 0.0
        for k in range(npts-1):
            ind = np.logical_and(zpts>=z_nodes[k], zpts<=z_nodes[k+1]) 
            section_mass[k] += rho_ballast * np.pi * np.trapz(R_id[ind]**2, zpts[ind])

        Ixx = Iyy = frustum.frustumIxx(R_id[:-1], R_id[1:], np.diff(zpts))
        Izz = frustum.frustumIzz(R_id[:-1], R_id[1:], np.diff(zpts))
        V_slice = frustum.frustumVol(R_id[:-1], R_id[1:], np.diff(zpts))
        I_keel = np.zeros((3,3))
        dz  = frustum.frustumCG(R_id[:-1], R_id[1:], np.diff(zpts)) + zpts[:-1] - z_draft
        for k in range(V_slice.size):
            R = np.array([0.0, 0.0, dz[k]])
            Icg = assembleI( [Ixx[k], Iyy[k], Izz[k], 0.0, 0.0, 0.0] )
            I_keel += Icg + V_slice[k]*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        I_keel = rho_ballast * unassembleI(I_keel)
        
        # Water ballast will start at top of fixed ballast
        z_water_start = (z_draft + h_ballast)
        #z_water_start = z_water_start + inputs['variable_ballast_start'] * (z_nodes[-1] - z_water_start)
        
        # Find height of water ballast numerically by finding the height that integrates to the mass we want
        # This step is completed in column.py or semi.py because we must account for other substructure elements too
        zpts    = np.linspace(z_water_start, 0.0, npts)
        R_id    = np.interp(zpts, z_nodes, R_id_orig)
        outputs['variable_ballast_interp_zpts']   = zpts
        outputs['variable_ballast_interp_radius'] = R_id
        
        # Save permanent ballast mass and variable height
        outputs['ballast_mass']   = section_mass
        outputs['ballast_I_keel'] = I_keel
        outputs['ballast_z_cg']   = z_cg_perm
        outputs['ballast_cost']   = inputs['ballast_cost_rate'] * m_perm


        
        
class ColumnGeometry(ExplicitComponent):
    """
    OpenMDAO Component class for vertical columns in substructure for floating offshore wind turbines.
    """

    def initialize(self):
        self.options.declare('nSection')
        self.options.declare('nFull')
        
    def setup(self):
        nSection = self.options['nSection']
        nFull    = self.options['nFull']

        # Design variables
        self.add_input('water_depth', val=0.0, units='m', desc='water depth')
        self.add_input('Hs', val=0.0, units='m', desc='significant wave height')
        self.add_input('freeboard', val=0.0, units='m', desc='Length of column above water line')
        self.add_input('max_draft', val=0.0, units='m', desc='Maxmimum length of column below water line')
        self.add_input('z_full_in', val=np.zeros((nFull,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('z_param_in', val=np.zeros((nSection+1,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('section_center_of_mass', val=np.zeros(nFull-1), units='m', desc='z position of center of mass of each can in the cylinder')

        self.add_input('stiffener_web_height', val=np.zeros((nSection,)), units='m', desc='height of stiffener web (base of T) within each section bottom to top (length = nsection)')
        self.add_input('stiffener_web_thickness', val=np.zeros((nSection,)), units='m', desc='thickness of stiffener web (base of T) within each section bottom to top (length = nsection)')
        self.add_input('stiffener_flange_width', val=np.zeros((nSection,)), units='m', desc='height of stiffener flange (top of T) within each section bottom to top (length = nsection)')
        self.add_input('stiffener_flange_thickness', val=np.zeros((nSection,)), units='m', desc='thickness of stiffener flange (top of T) within each section bottom to top (length = nsection)')
        self.add_input('stiffener_spacing', val=np.zeros((nSection,)), units='m', desc='Axial distance from one ring stiffener to another within each section bottom to top (length = nsection)')

        # Outputs
        self.add_output('z_full', val=np.zeros((nFull,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_output('z_param', val=np.zeros((nSection+1,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_output('draft', val=0.0, units='m', desc='Column draft (length of body under water)')
        self.add_output('z_section', val=np.zeros((nFull-1,)), units='m', desc='z-coordinates of section centers of mass (length = nsection)')


        self.add_output('h_web', val=np.zeros((nFull-1,)), units='m', desc='height of stiffener web (base of T) within each section bottom to top')
        self.add_output('t_web', val=np.zeros((nFull-1,)), units='m', desc='thickness of stiffener web (base of T) within each section bottom to top')
        self.add_output('w_flange', val=np.zeros((nFull-1,)), units='m', desc='height of stiffener flange (top of T) within each section bottom to top')
        self.add_output('t_flange', val=np.zeros((nFull-1,)), units='m', desc='thickness of stiffener flange (top of T) within each section bottom to top')
        self.add_output('L_stiffener', val=np.zeros((nFull-1,)), units='m', desc='Axial distance from one ring stiffener to another within each section bottom to top')
        
        # Output constraints
        self.add_output('draft_margin', val=0.0, desc='Ratio of draft to water depth')
        self.add_output('wave_height_freeboard_ratio', val=0.0, desc='Ratio of maximum wave height (avg of top 1%) to freeboard')

        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):
        # Unpack variables
        freeboard = inputs['freeboard']

        # With waterline at z=0, set the z-position of section nodes
        # Note sections and nodes start at bottom of column and move up
        draft     = inputs['z_param_in'][-1] - freeboard
        z_full    = inputs['z_full_in'] - draft 
        z_param   = inputs['z_param_in'] - draft 
        z_section = inputs['section_center_of_mass'] - draft 
        outputs['draft']     = draft
        outputs['z_full']    = z_full
        outputs['z_param']   = z_param
        outputs['z_section'] = z_section

        # Create constraint output that draft is less than water depth
        outputs['draft_margin'] = draft / inputs['max_draft']

        # Make sure freeboard is more than 20% of Hs (DNV-OS-J101)
        outputs['wave_height_freeboard_ratio'] = inputs['Hs'] / np.abs(freeboard)

        # Sectional stiffener properties
        outputs['t_web']        = sectionalInterp(z_section, z_param, inputs['stiffener_web_thickness'])
        outputs['t_flange']     = sectionalInterp(z_section, z_param, inputs['stiffener_flange_thickness'])
        outputs['h_web']        = sectionalInterp(z_section, z_param, inputs['stiffener_web_height'])
        outputs['w_flange']     = sectionalInterp(z_section, z_param, inputs['stiffener_flange_width'])
        outputs['L_stiffener']  = sectionalInterp(z_section, z_param, inputs['stiffener_spacing'])
        


class ColumnProperties(ExplicitComponent):
    """
    OpenMDAO Component class for column substructure elements in floating offshore wind turbines.
    """

    def initialize(self):
        self.options.declare('nFull')
        
    def setup(self):
        nFull    = self.options['nFull']

        # Variables local to the class and not OpenMDAO
        self.ibox = None
        
        # Environment
        self.add_input('water_density', val=0.0, units='kg/m**3', desc='density of water')

        # Inputs from Geometry
        self.add_input('z_full', val=np.zeros((nFull,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('z_section', val=np.zeros((nFull-1,)), units='m', desc='z-coordinates of section centers of mass (length = nsection)')

        # Design variables
        self.add_input('d_full', val=np.zeros((nFull,)), units='m', desc='outer diameter at each section node bottom to top (length = nsection + 1)')
        self.add_input('t_full', val=np.zeros((nFull-1,)), units='m', desc='shell wall thickness at each section node bottom to top (length = nsection + 1)')
        self.add_input('buoyancy_tank_diameter', val=0.0, units='m', desc='Radius of heave plate at bottom of column')
        
        # Mass correction factors from simple rules here to real life
        self.add_input('shell_mass', val=np.zeros(nFull-1), units='kg', desc='mass of column shell')
        self.add_input('stiffener_mass', val=np.zeros(nFull-1), units='kg', desc='mass of column stiffeners')
        self.add_input('bulkhead_mass', val=np.zeros(nFull), units='kg', desc='mass of column bulkheads')
        self.add_input('buoyancy_tank_mass', val=0.0, units='kg', desc='mass of heave plate')
        self.add_input('ballast_mass', val=np.zeros(nFull-1), units='kg', desc='mass of permanent ballast')

        self.add_input('buoyancy_tank_cg', val=0.0, units='m', desc='z-coordinate of center of mass for buoyancy tank')
        self.add_input('ballast_z_cg', val=0.0, units='m', desc='z-coordinate or permanent ballast center of gravity')
        self.add_input('column_mass_factor', val=0.0, desc='Overall column mass correction factor')
        self.add_input('outfitting_mass_fraction', val=0.0, desc='Mass fraction added for outfitting')

        # Moments of inertia
        self.add_input('shell_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of outer shell relative to keel point')
        self.add_input('bulkhead_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of bulkheads relative to keel point')
        self.add_input('stiffener_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of stiffeners relative to keel point')
        self.add_input('buoyancy_tank_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of heave plate relative to keel point')
        self.add_input('ballast_I_keel', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of permanent ballast relative to keel point')

        # For buoyancy
        self.add_input('buoyancy_tank_displacement', val=0.0, units='m**3', desc='volume of water displaced by buoyancy tank')
        
        # Costs and cost rates
        self.add_input('shell_cost', val=0.0, units='USD', desc='mass of column shell')
        self.add_input('stiffener_cost', val=0.0, units='USD', desc='mass of column stiffeners')
        self.add_input('bulkhead_cost', val=0.0, units='USD', desc='mass of column bulkheads')
        self.add_input('ballast_cost', val=0.0, units='USD', desc='cost of permanent ballast')
        self.add_input('buoyancy_tank_cost', val=0.0, units='USD', desc='mass of heave plate')
        self.add_input('material_cost_rate', 0.0, units='USD/kg', desc='Raw material cost: steel $1.1/kg, aluminum $3.5/kg')
        self.add_input('outfitting_cost_rate', val=0.0, units='USD/kg', desc='Cost per unit mass for outfitting column')

        # Outputs
        self.add_output('z_center_of_mass', val=0.0, units='m', desc='z-position CofG of column')
        self.add_output('z_center_of_buoyancy', val=0.0, units='m', desc='z-position CofB of column')
        self.add_output('Awater', val=0.0, units='m**2', desc='Area of waterplace cross section')
        self.add_output('Iwater', val=0.0, units='m**4', desc='Second moment of area of waterplace cross section')
        self.add_output('I_column', val=np.zeros(6), units='kg*m**2', desc='Moments of inertia of whole column relative to keel point')
        self.add_output('displaced_volume', val=np.zeros((nFull-1,)), units='m**3', desc='Volume of water displaced by column by section')
        self.add_output('hydrostatic_force', val=np.zeros((nFull-1,)), units='N', desc='Net z-force on column sections')
 
        self.add_output('column_structural_mass', val=0.0, units='kg', desc='mass of column structure')
        
        self.add_output('column_outfitting_cost', val=0.0, units='USD', desc='cost of outfitting the column')
        self.add_output('column_outfitting_mass', val=0.0, units='kg', desc='cost of outfitting the column')

        self.add_output('column_added_mass', val=np.zeros(6), units='kg', desc='hydrodynamic added mass matrix diagonal')
        self.add_output('column_total_mass', val=np.zeros((nFull-1,)), units='kg', desc='total mass of column by section')
        self.add_output('column_total_cost', val=0.0, units='USD', desc='total cost of column')
        self.add_output('column_structural_cost', val=0.0, units='USD', desc='Cost of column without ballast or outfitting')
        self.add_output('tapered_column_cost_rate', val=0.0, units='USD/t', desc='Cost rate of finished column')
        
        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
        
    def compute(self, inputs, outputs):

        self.compute_column_mass_cg(inputs, outputs)
        self.balance_column(inputs, outputs)
        self.compute_cost(inputs, outputs)

        

    def compute_column_mass_cg(self, inputs, outputs):
        """Computes column mass from components: Shell, Stiffener rings, Bulkheads
        Also computes center of mass of the shell by weighted sum of the components' position
        
        INPUTS:
        ----------
        inputs   : dictionary of input parameters
        outputs : dictionary of output parameters
        
        OUTPUTS:
        ----------
        section_mass class variable set
        m_column   : column mass
        z_cg     : center of mass along z-axis for the column
        column_mass       in 'outputs' dictionary set
        shell_mass      in 'outputs' dictionary set
        stiffener_mass  in 'outputs' dictionary set
        bulkhead_mass   in 'outputs' dictionary set
        outfitting_mass in 'outputs' dictionary set
        """
        # Unpack variables
        out_frac     = inputs['outfitting_mass_fraction']
        coeff        = inputs['column_mass_factor']
        z_nodes      = inputs['z_full']
        z_section    = inputs['z_section']
        z_box        = inputs['buoyancy_tank_cg']
        z_ballast    = inputs['ballast_z_cg']
        m_shell      = inputs['shell_mass']
        m_stiffener  = inputs['stiffener_mass']
        m_bulkhead   = inputs['bulkhead_mass']
        m_box        = inputs['buoyancy_tank_mass']
        m_ballast    = inputs['ballast_mass']
        I_shell      = inputs['shell_I_keel']
        I_stiffener  = inputs['stiffener_I_keel']
        I_bulkhead   = inputs['bulkhead_I_keel']
        I_box        = inputs['buoyancy_tank_I_keel']
        I_ballast    = inputs['ballast_I_keel']

        # Consistency check
        if out_frac > 1.0: out_frac -= 1.0
        
        # Initialize summations
        m_column  = 0.0
        z_cg      = 0.0
        
        # Find mass of all of the sub-components of the column
        # Masses assumed to be focused at section centroids
        m_column += (m_shell + m_stiffener).sum()
        z_cg     += np.dot(m_shell+m_stiffener, z_section)

        # Masses assumed to be centered at nodes
        m_column += m_bulkhead.sum()
        z_cg     += np.dot(m_bulkhead, z_nodes)

        # Mass with variable location
        m_column += m_box
        z_cg     += m_box*z_box

        # Account for components not explicitly calculated here
        m_column *= coeff

        # Compute CG position of the column
        z_cg     *= coeff / m_column

        # Now calculate outfitting mass, evenly distributed so cg doesn't change
        m_outfit  = out_frac * m_column

        # Add in ballast
        m_total   = m_column + m_outfit + m_ballast.sum()
        z_cg      = ( (m_column+m_outfit)*z_cg + m_ballast.sum()*z_ballast ) / m_total

        # Find sections for ballast and buoyancy tank
        ibox  = 0
        try:
            ibox  = np.where(z_box >= z_nodes)[0][-1]
        except:
            print(z_box, z_ballast, z_nodes)
        self.ibox = ibox

        # Now do tally by section
        m_sections         = coeff*(m_shell + m_stiffener + m_bulkhead[:-1]) + m_ballast
        m_sections        += m_outfit / m_shell.size
        m_sections[-1]    += coeff*m_bulkhead[-1]
        m_sections[ibox]  += coeff*m_box

        # Add up moments of inertia at keel, make sure to scale mass appropriately
        I_total   = ((1+out_frac) * coeff) * (I_shell + I_stiffener + I_bulkhead + I_box) + I_ballast

        # Move moments of inertia from keel to cg
        I_total  -= m_total*((z_cg-z_nodes[0])**2.0) * np.r_[1.0, 1.0, np.zeros(4)]
        I_total   = np.maximum(I_total, 0.0)

        # Store outputs addressed so far
        outputs['column_total_mass']      = m_sections
        outputs['column_structural_mass'] = m_column + m_outfit
        outputs['column_outfitting_mass'] = m_outfit
        outputs['z_center_of_mass']       = z_cg
        outputs['I_column']               = I_total
        

        
    def balance_column(self, inputs, outputs):
        # Unpack variables
        R_od              = 0.5*inputs['d_full']
        R_plate           = 0.5*inputs['buoyancy_tank_diameter']
        z_nodes           = inputs['z_full']
        z_box             = inputs['buoyancy_tank_cg']
        V_box             = inputs['buoyancy_tank_displacement']
        rho_water         = inputs['water_density']
        nsection          = R_od.size - 1

        # Compute volume of each section and mass of displaced water by section
        # Find the radius at the waterline so that we can compute the submerged volume as a sum of frustum sections
        if z_nodes[-1] > 0.0:
            r_waterline = np.interp(0.0, z_nodes, R_od)
            z_under     = np.r_[z_nodes[z_nodes < 0.0], 0.0]
            r_under     = np.r_[R_od[z_nodes < 0.0], r_waterline]
        else:
            r_waterline = R_od[-1]
            r_under     = R_od
            z_under     = z_nodes

        # Submerged volume (with zero-padding)
        V_under = frustum.frustumVol(r_under[:-1], r_under[1:], np.diff(z_under))
        add0    = np.maximum(0, nsection-V_under.size)
        outputs['displaced_volume'] = np.r_[V_under, np.zeros(add0)]
        outputs['displaced_volume'][self.ibox] += V_box

        # Compute Center of Buoyancy in z-coordinates (0=waterline)
        # First get z-coordinates of CG of all frustums
        z_cg_under  = frustum.frustumCG(r_under[:-1], r_under[1:], np.diff(z_under))
        z_cg_under += z_under[:-1]
        # Now take weighted average of these CG points with volume
        z_cb     = ( (V_box*z_box) + np.dot(V_under, z_cg_under)) / (outputs['displaced_volume'].sum() + eps)
        outputs['z_center_of_buoyancy'] = z_cb

        # Find total hydrostatic force by section- sign says in which direction force acts
        # Since we are working on z_under grid, need to redefine z_section, ibox, etc.
        z_undersec,_ = nodal2sectional(z_under)
        if z_box > 0.0 and V_box == 0.0:
            ibox = 0
        else:
            ibox = np.where(z_box >= z_under)[0][-1]
        F_hydro      = np.pi * np.diff(r_under**2.0) * np.maximum(0.0, -z_undersec) #cg_under))
        if F_hydro.size > 0:
            F_hydro[0] += np.pi * r_under[0]**2 * (-z_under[0])
            if z_nodes[-1] < 0.0:
                F_hydro[-1] -= np.pi * r_under[-1]**2 * (-z_under[-1])
            F_hydro[ibox] += V_box
            F_hydro    *= rho_water * gravity
        outputs['hydrostatic_force'] = np.r_[F_hydro, np.zeros(add0)]
        
        # 2nd moment of area for circular cross section
        # Note: Assuming Iwater here depends on "water displacement" cross-section
        # and not actual moment of inertia type of cross section (thin hoop)
        outputs['Iwater'] = 0.25 * np.pi * r_waterline**4.0
        outputs['Awater'] = np.pi * r_waterline**2.0

        # Calculate diagonal entries of added mass matrix
        # Prep for integrals too
        npts     = 100 * R_od.size
        zpts     = np.linspace(z_under[0], z_under[-1], npts)
        r_under  = np.interp(zpts, z_under, r_under)
        m_a      = np.zeros(6)
        m_a[:2]  = rho_water * outputs['displaced_volume'].sum() # A11 surge, A22 sway
        m_a[2]   = 0.5 * (8.0/3.0) * rho_water * np.maximum(R_plate, r_under.max())**3.0# A33 heave
        m_a[3:5] = np.pi * rho_water * np.trapz((zpts-z_cb)**2.0 * r_under**2.0, zpts)# A44 roll, A55 pitch
        m_a[5]   = 0.0 # A66 yaw
        outputs['column_added_mass'] = m_a

        
    def compute_cost(self, inputs, outputs):
        outputs['column_structural_cost']   = inputs['column_mass_factor']*(inputs['shell_cost'] + inputs['stiffener_cost'] +
                                                                             inputs['bulkhead_cost'] + inputs['buoyancy_tank_cost'])
        outputs['column_outfitting_cost']   = inputs['outfitting_cost_rate'] * outputs['column_outfitting_mass']
        outputs['column_total_cost']        = outputs['column_structural_cost'] + outputs['column_outfitting_cost'] + inputs['ballast_cost']
        outputs['tapered_column_cost_rate'] = 1e3*outputs['column_total_cost']/outputs['column_total_mass'].sum()

        
class ColumnBuckling(ExplicitComponent):
    '''
    This function computes the applied axial and hoop stresses in a column and compares that to 
    limits established by the API standard.  Some physcial geometry checks are also performed.
    '''

    def initialize(self):
        self.options.declare('nSection')
        self.options.declare('nFull')
        
    def setup(self):
        nSection = self.options['nSection']
        nFull    = self.options['nFull']

        # From other modules
        self.add_input('stack_mass_in', val=eps, units='kg', desc='Weight above the cylinder column')
        self.add_input('section_mass', val=np.zeros((nFull-1,)), units='kg', desc='total mass of column by section')
        self.add_input('pressure', np.zeros(nFull), units='N/m**2', desc='Dynamic (and static)? pressure')
        
        self.add_input('d_full', np.zeros(nFull), units='m', desc='cylinder diameter at corresponding locations')
        self.add_input('t_full', np.zeros(nFull-1), units='m', desc='shell thickness at corresponding locations')
        self.add_input('z_full', val=np.zeros(nFull), units='m', desc='z-coordinates of section nodes (length = nsection+1)')

        self.add_input('h_web', val=np.zeros((nFull-1,)), units='m', desc='height of stiffener web (base of T) within each section bottom to top')
        self.add_input('t_web', val=np.zeros((nFull-1,)), units='m', desc='thickness of stiffener web (base of T) within each section bottom to top')
        self.add_input('w_flange', val=np.zeros((nFull-1,)), units='m', desc='height of stiffener flange (top of T) within each section bottom to top')
        self.add_input('t_flange', val=np.zeros((nFull-1,)), units='m', desc='thickness of stiffener flange (top of T) within each section bottom to top')
        self.add_input('L_stiffener', val=np.zeros((nFull-1,)), units='m', desc='Axial distance from one ring stiffener to another within each section bottom to top')

        self.add_input('E', val=0.0, units='Pa', desc='Modulus of elasticity (Youngs) of material')
        self.add_input('nu', val=0.0, desc='poissons ratio of column material')
        self.add_input('yield_stress', val=0.0, units='Pa', desc='yield stress of material')

        self.add_discrete_input('loading', val='hydro', desc='Loading type in API checks [hydro/radial]')
        self.add_input('gamma_f', 0.0, desc='safety factor on loads')
        self.add_input('gamma_b', 0.0, desc='buckling safety factor')
        
        # Output constraints
        self.add_output('flange_compactness', val=np.zeros((nFull-1,)), desc='check for flange compactness')
        self.add_output('web_compactness', val=np.zeros((nFull-1,)), desc='check for web compactness')
        
        self.add_output('axial_local_api', val=np.zeros((nFull-1,)), desc='unity check for axial load with API safety factors - local buckling')
        self.add_output('axial_general_api', val=np.zeros((nFull-1,)), desc='unity check for axial load with API safety factors- genenral instability')
        self.add_output('external_local_api', val=np.zeros((nFull-1,)), desc='unity check for external pressure with API safety factors- local buckling')
        self.add_output('external_general_api', val=np.zeros((nFull-1,)), desc='unity check for external pressure with API safety factors- general instability')

        self.add_output('axial_local_utilization', val=np.zeros((nFull-1,)), desc='utilization check for axial load - local buckling')
        self.add_output('axial_general_utilization', val=np.zeros((nFull-1,)), desc='utilization check for axial load - genenral instability')
        self.add_output('external_local_utilization', val=np.zeros((nFull-1,)), desc='utilization check for external pressure - local buckling')
        self.add_output('external_general_utilization', val=np.zeros((nFull-1,)), desc='utilization check for external pressure - general instability')
        
        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute_applied_axial(self, inputs):
        """Compute axial stress for column from z-axis loading

        INPUTS:
        ----------
        inputs       : dictionary of input parameters
        section_mass : float (scalar/vector),  mass of each column section as axial loading increases with column depth

        OUTPUTS:
        -------
        stress   : float (scalar/vector),  axial stress
        """
        # Unpack variables
        R_od,_         = nodal2sectional(inputs['d_full'])
        R_od          *= 0.5
        t_wall         = inputs['t_full']
        section_mass   = inputs['section_mass']
        m_stack        = inputs['stack_mass_in']
        
        # Middle radius
        R_m = R_od - 0.5*t_wall
        # Add in weight of sections above it
        axial_load = m_stack + np.r_[0.0, np.cumsum(section_mass[:-1])]
        # Divide by shell cross sectional area to get stress
        return (gravity * axial_load / (2.0 * np.pi * R_m * t_wall))

    
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack variables
        R_od,_       = nodal2sectional( inputs['d_full'] )
        R_od        *= 0.5
        h_section    = np.diff( inputs['z_full'] )
        t_wall       = inputs['t_full']
        
        t_web        = inputs['t_web']
        t_flange     = inputs['t_flange']
        h_web        = inputs['h_web']
        w_flange     = inputs['w_flange']
        L_stiffener  = inputs['L_stiffener']

        gamma_f      = inputs['gamma_f']
        gamma_b      = inputs['gamma_b']
        
        E            = inputs['E'] # Young's modulus
        nu           = inputs['nu'] # Poisson ratio
        sigma_y      = inputs['yield_stress']
        loading      = discrete_inputs['loading']
        nodalP,_     = nodal2sectional( inputs['pressure'] )
        pressure     = 1e-12 if loading in ['ax','axial','testing','test'] else nodalP+1e-12

        # Apply quick "compactness" check on stiffener geometry
        # Constraint is that these must be >= 1
        flange_compactness = 0.375 * (t_flange / (0.5*w_flange)) * np.sqrt(E / sigma_y)
        web_compactness    = 1.0   * (t_web    / h_web         ) * np.sqrt(E / sigma_y)

        # Compute applied axial stress simply, like API guidelines (as opposed to running frame3dd)
        sigma_ax = self.compute_applied_axial(inputs)
        (axial_local_api, axial_general_api, external_local_api, external_general_api,
         axial_local_raw, axial_general_raw, external_local_raw, external_general_raw) = shellBuckling_withStiffeners(
             pressure, sigma_ax, R_od, t_wall, h_section,
             h_web, t_web, w_flange, t_flange,
             L_stiffener, E, nu, sigma_y, loading)
        
        outputs['flange_compactness']     = flange_compactness
        outputs['web_compactness']        = web_compactness
        
        outputs['axial_local_api']      = axial_local_api
        outputs['axial_general_api']    = axial_general_api
        outputs['external_local_api']   = external_local_api
        outputs['external_general_api'] = external_general_api

        outputs['axial_local_utilization']      = axial_local_raw * gamma_f*gamma_b
        outputs['axial_general_utilization']    = axial_general_raw * gamma_f*gamma_b
        outputs['external_local_utilization']   = external_local_raw * gamma_f*gamma_b
        outputs['external_general_utilization'] = external_general_raw * gamma_f*gamma_b


class Column(Group):

    def initialize(self):
        self.options.declare('nSection')
        self.options.declare('nFull')
        self.options.declare('topLevelFlag', default=False)
        
    def setup(self):
        nSection = self.options['nSection']
        nFull    = self.options['nFull']
        nRefine  = (nFull-1)/nSection
        topLevelFlag = self.options['topLevelFlag']

        columnIndeps = IndepVarComp()
        columnIndeps.add_output('freeboard', 0.0, units='m')
        columnIndeps.add_output('section_height', np.zeros((nSection,)), units='m')
        columnIndeps.add_output('outer_diameter', np.zeros((nSection+1,)), units='m')
        columnIndeps.add_output('wall_thickness', np.zeros((nSection,)), units='m')
        columnIndeps.add_output('stiffener_web_height', np.zeros((nSection,)), units='m')
        columnIndeps.add_output('stiffener_web_thickness', np.zeros((nSection,)), units='m')
        columnIndeps.add_output('stiffener_flange_width', np.zeros((nSection,)), units='m')
        columnIndeps.add_output('stiffener_flange_thickness', np.zeros((nSection,)), units='m')
        columnIndeps.add_output('stiffener_spacing', np.zeros((nSection,)), units='m')
        columnIndeps.add_output('bulkhead_thickness', np.zeros((nSection+1,)), units='m')
        columnIndeps.add_output('permanent_ballast_height', 0.0, units='m')
        columnIndeps.add_output('buoyancy_tank_diameter', 0.0, units='m')
        columnIndeps.add_output('buoyancy_tank_height', 0.0, units='m')
        columnIndeps.add_output('buoyancy_tank_location', 0.0, units='m')
        columnIndeps.add_output('material_density', 0.0, units='kg/m**3')
        self.add_subsystem('columnIndeps', columnIndeps, promotes=['*'])

        if topLevelFlag:
            sharedIndeps = IndepVarComp()
            sharedIndeps.add_output('water_density', 0.0, units='kg/m**3')
            sharedIndeps.add_output('water_depth', 0.0, units='m')
            sharedIndeps.add_output('wave_beta', 0.0, units='deg')
            sharedIndeps.add_output('wave_z0', 0.0, units='m')
            sharedIndeps.add_output('significant_wave_height', 0.0, units='m')
            self.add_subsystem('sharedIndeps', sharedIndeps, promotes=['*'])
            
        
        self.add_subsystem('cyl_geom', CylinderDiscretization(nPoints=nSection+1, nRefine=nRefine), promotes=['section_height','diameter','wall_thickness',
                                                                                    'd_full','t_full','foundation_height'])
        
        self.add_subsystem('gc', GeometricConstraints(nPoints=nSection+1, diamFlag=True), promotes=['max_taper','min_d_to_t','manufacturability','weldability'])

        self.add_subsystem('cyl_mass', CylinderMass(nPoints=nFull), promotes=['d_full','t_full','material_density',
                                                            'material_cost_rate','labor_cost_rate','painting_cost_rate',
                                                            'section_center_of_mass'])

        self.add_subsystem('col_geom', ColumnGeometry(nSection=nSection, nFull=nFull), promotes=['*'])

        self.add_subsystem('bulk', BulkheadProperties(nSection=nSection, nFull=nFull), promotes=['*'])

        self.add_subsystem('stiff', StiffenerProperties(nSection=nSection, nFull=nFull), promotes=['*'])

        self.add_subsystem('plate', BuoyancyTankProperties(nFull=nFull), promotes=['*'])

        self.add_subsystem('ball', BallastProperties(nFull=nFull), promotes=['*'])

        self.add_subsystem('col', ColumnProperties(nFull=nFull), promotes=['*'])

        self.add_subsystem('wind', PowerWind(nPoints=nFull), promotes=['Uref','zref','shearExp','z0'])
        self.add_subsystem('wave', LinearWaves(nPoints=nFull), promotes=['Uc','hmax','T'])
        self.add_subsystem('windLoads', CylinderWindDrag(nPoints=nFull), promotes=['cd_usr','beta'])
        self.add_subsystem('waveLoads', CylinderWaveDrag(nPoints=nFull), promotes=['cm','cd_usr'])
        self.add_subsystem('distLoads', AeroHydroLoads(nPoints=nFull), promotes=['Px','Py','Pz','qdyn','yaw'])

        self.add_subsystem('buck', ColumnBuckling(nSection=nSection, nFull=nFull), promotes=['*'])

        self.connect('outer_diameter', ['diameter', 'gc.d'])
        self.connect('wall_thickness', 'gc.t')
        self.connect('cyl_geom.z_param', 'z_param_in')
        self.connect('cyl_geom.z_full', ['cyl_mass.z_full','z_full_in'])
        
        #self.connect('cyl_mass.section_center_of_mass', 'col_geom.section_center_of_mass')
        
        self.connect('cyl_mass.mass', 'shell_mass')
        self.connect('cyl_mass.cost', 'shell_cost')
        self.connect('cyl_mass.I_base', 'shell_I_keel')
        self.connect('material_density','rho')
        
        self.connect('column_total_mass', 'section_mass')

        if topLevelFlag:
            self.connect('water_depth','wave.z_floor')
            self.connect('wave_beta','waveLoads.beta')
            self.connect('wave_z0', 'wave.z_surface')
            self.connect('significant_wave_height',['Hs', 'hmax'])
            self.connect('water_density',['wave.rho','waveLoads.rho'])
        self.connect('z_full', ['wind.z', 'wave.z', 'windLoads.z','waveLoads.z','distLoads.z'])
        self.connect('d_full', ['windLoads.d','waveLoads.d'])

        self.connect('wind.U', 'windLoads.U')

        self.connect('wave.U', 'waveLoads.U')
        self.connect('wave.A', 'waveLoads.A')
        self.connect('wave.p', 'waveLoads.p')
        
        # connections to distLoads1
        self.connect('windLoads.windLoads_Px', 'distLoads.windLoads_Px')
        self.connect('windLoads.windLoads_Py', 'distLoads.windLoads_Py')
        self.connect('windLoads.windLoads_Pz', 'distLoads.windLoads_Pz')
        self.connect('windLoads.windLoads_qdyn', 'distLoads.windLoads_qdyn')
        self.connect('windLoads.windLoads_beta', 'distLoads.windLoads_beta')
        self.connect('windLoads.windLoads_z', 'distLoads.windLoads_z')
        self.connect('windLoads.windLoads_d', 'distLoads.windLoads_d')
        
        self.connect('waveLoads.waveLoads_Px', 'distLoads.waveLoads_Px')
        self.connect('waveLoads.waveLoads_Py', 'distLoads.waveLoads_Py')
        self.connect('waveLoads.waveLoads_Pz', 'distLoads.waveLoads_Pz')
        self.connect('waveLoads.waveLoads_pt', 'distLoads.waveLoads_qdyn')
        self.connect('waveLoads.waveLoads_beta', 'distLoads.waveLoads_beta')
        self.connect('waveLoads.waveLoads_z', 'distLoads.waveLoads_z')
        self.connect('waveLoads.waveLoads_d', 'distLoads.waveLoads_d')

        self.connect('qdyn', 'pressure')



"""
tcc_csm_component.py

Created by NWTC Systems Engineering Sub-Task on 2012-08-01.
Copyright (c) NREL. All rights reserved.
"""

from __future__ import print_function
import numpy as np

from openmdao.api import ExplicitComponent, Problem, Group, IndepVarComp

from wisdem.turbine_costsse.turbine_costsse_2015 import Turbine_CostsSE_2015

# --------------------------------------------------------------------
class BladeMass(ExplicitComponent):
    
    def setup(self):
        
        # Variables
        self.add_input('rotor_diameter', 0.0, units='m', desc= 'rotor diameter of the machine')
        self.add_discrete_input('turbine_class', 1, desc='turbine class')
        self.add_discrete_input('blade_has_carbon', False, desc= 'does the blade have carbon?') #default to doesn't have carbon
        self.add_input('blade_mass_coeff', 0.5, desc= 'A in the blade mass equation: A*(rotor_diameter/B)^exp') #default from ppt
        self.add_input('blade_user_exp', 2.5, desc='optional user-entered exp for the blade mass equation')
        
        # Outputs
        self.add_output('blade_mass', 0.0, units='kg', desc= 'component mass [kg]')
  
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        
        rotor_diameter = inputs['rotor_diameter']
        turbine_class = discrete_inputs['turbine_class']
        blade_has_carbon = discrete_inputs['blade_has_carbon']
        blade_mass_coeff = inputs['blade_mass_coeff']
        blade_user_exp = inputs['blade_user_exp']
    
        # select the exp for the blade mass equation
        exp = 0.0
        if turbine_class == 1:
            if blade_has_carbon:
              exp = 2.47
            else:
              exp = 2.54
        elif turbine_class > 1:
            if blade_has_carbon:
              exp = 2.44
            else:
              exp = 2.50
        else:
            exp = blade_user_exp
        
        # calculate the blade mass
        outputs['blade_mass'] = blade_mass_coeff * (rotor_diameter / 2)**exp

  # --------------------------------------------------------------------
class HubMass(ExplicitComponent):

    def setup(self):
        
        # Variables
        self.add_input('blade_mass', 0.0, units='kg', desc= 'component mass [kg]')
        self.add_input('hub_mass_coeff', 2.3, desc= 'A in the hub mass equation: A*blade_mass + B') #default from ppt
        self.add_input('hub_mass_intercept', 1320., desc= 'B in the hub mass equation: A*blade_mass + B') #default from ppt
        
        # Outputs
        self.add_output('hub_mass', 0.0, units='kg', desc='component mass [kg]')
  
    def compute(self, inputs, outputs):
      
        blade_mass = inputs['blade_mass']
        hub_mass_coeff = inputs['hub_mass_coeff']
        hub_mass_intercept = inputs['hub_mass_intercept']
        
        # calculate the hub mass
        outputs['hub_mass'] = hub_mass_coeff * blade_mass + hub_mass_intercept

# --------------------------------------------------------------------
class PitchSystemMass(ExplicitComponent):
    
    def setup(self):
        
        self.add_input('blade_mass', 0.0, units='kg', desc= 'component mass [kg]')
        self.add_discrete_input('blade_number', 3, desc='number of rotor blades')
        self.add_input('pitch_bearing_mass_coeff', 0.1295, desc='A in the pitch bearing mass equation: A*blade_mass*blade_number + B') #default from old CSM
        self.add_input('pitch_bearing_mass_intercept', 491.31, desc='B in the pitch bearing mass equation: A*blade_mass*blade_number + B') #default from old CSM
        self.add_input('bearing_housing_percent', .3280, desc='bearing housing percentage (in decimal form: ex 10% is 0.10)') #default from old CSM
        self.add_input('mass_sys_offset', 555.0, desc='mass system offset') #default from old CSM
        
        # Outputs
        self.add_output('pitch_system_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        
        blade_mass = inputs['blade_mass']
        blade_number = discrete_inputs['blade_number']
        pitch_bearing_mass_coeff = inputs['pitch_bearing_mass_coeff']
        pitch_bearing_mass_intercept = inputs['pitch_bearing_mass_intercept']
        bearing_housing_percent = inputs['bearing_housing_percent']
        mass_sys_offset = inputs['mass_sys_offset']
        
        # calculate the hub mass
        pitchBearingMass = pitch_bearing_mass_coeff * blade_mass * blade_number + pitch_bearing_mass_intercept
        outputs['pitch_system_mass'] = pitchBearingMass * (1 + bearing_housing_percent) + mass_sys_offset

# --------------------------------------------------------------------
class SpinnerMass(ExplicitComponent):

    def setup(self):
        
    
        # Variables
        self.add_input('rotor_diameter', 0.0, units='m', desc= 'rotor diameter of the machine')
        self.add_input('spinner_mass_coeff', 15.5, desc= 'A in the spinner mass equation: A*rotor_diameter + B')
        self.add_input('spinner_mass_intercept', -980.0, desc= 'B in the spinner mass equation: A*rotor_diameter + B')
        
        # Outputs
        self.add_output('spinner_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        rotor_diameter = inputs['rotor_diameter']
        spinner_mass_coeff = inputs['spinner_mass_coeff']
        spinner_mass_intercept = inputs['spinner_mass_intercept']
        
        # calculate the spinner mass
        outputs['spinner_mass'] = spinner_mass_coeff * rotor_diameter + spinner_mass_intercept

# --------------------------------------------------------------------
class LowSpeedShaftMass(ExplicitComponent):

    def setup(self):
        
        
        # Variables
        self.add_input('blade_mass', 0.0, units='kg', desc='mass for a single wind turbine blade')
        self.add_input('machine_rating', 0.0, units='kW', desc='machine rating')
        self.add_input('lss_mass_coeff', 13., desc='A in the lss mass equation: A*(blade_mass*rated_power)^exp + B')
        self.add_input('lss_mass_exp', 0.65, desc='exp in the lss mass equation: A*(blade_mass*rated_power)^exp + B')
        self.add_input('lss_mass_intercept', 775., desc='B in the lss mass equation: A*(blade_mass*rated_power)^exp + B')
        
        # Outputs
        self.add_output('lss_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        blade_mass = inputs['blade_mass']
        machine_rating = inputs['machine_rating']
        lss_mass_coeff = inputs['lss_mass_coeff']
        lss_mass_exp = inputs['lss_mass_exp']
        lss_mass_intercept = inputs['lss_mass_intercept']
    
        # calculate the lss mass
        outputs['lss_mass'] = lss_mass_coeff * (blade_mass * machine_rating/1000.)**lss_mass_exp + lss_mass_intercept

# --------------------------------------------------------------------
class BearingMass(ExplicitComponent):

    def setup(self):


        # Variables
        self.add_input('rotor_diameter', 0.0, units='m', desc= 'rotor diameter of the machine')
        self.add_input('bearing_mass_coeff', 0.0001, desc= 'A in the bearing mass equation: A*rotor_diameter^exp') #default from ppt
        self.add_input('bearing_mass_exp', 3.5, desc= 'exp in the bearing mass equation: A*rotor_diameter^exp') #default from ppt
        
        # Outputs
        self.add_output('main_bearing_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        rotor_diameter = inputs['rotor_diameter']
        bearing_mass_coeff = inputs['bearing_mass_coeff']
        bearing_mass_exp = inputs['bearing_mass_exp']
        
        # calculates the mass of a SINGLE bearing
        outputs['main_bearing_mass'] = bearing_mass_coeff * rotor_diameter ** bearing_mass_exp

# --------------------------------------------------------------------
class RotorTorque(ExplicitComponent):

    def setup(self):
  
        # Variables
        self.add_input('rotor_diameter', 0.0, units='m', desc= 'rotor diameter of the machine')
        self.add_input('machine_rating', 0.0, units='kW', desc='machine rating')
        self.add_input('max_tip_speed', 0.0, units='m/s', desc='Maximum allowable blade tip speed')
        self.add_input('max_efficiency', 0.0, desc='Maximum possible drivetrain efficiency')

        self.add_output('rated_rpm', 0.0, units='rpm', desc = 'rpm of rotor at rated power') #JMF do we want this default?
        self.add_output('rotor_torque', 0.0, units='N*m', desc = 'torque from rotor at rated power') #JMF do we want this default?

    def compute(self, inputs, outputs):
        # Rotor force calculations for nacelle inputs
        maxTipSpd     = inputs['max_tip_speed']
        maxEfficiency = inputs['max_efficiency']
        
        ratedHubPower_W  = inputs['machine_rating']*1000. / maxEfficiency 
        rotorSpeed       = maxTipSpd / (0.5*inputs['rotor_diameter'])
        outputs['rated_rpm']    = rotorSpeed / (2*np.pi) * 60.0
        outputs['rotor_torque'] = ratedHubPower_W / rotorSpeed
# --------------------------------------------------------------------
class GearboxMass(ExplicitComponent):

    def setup(self):
  
  
        # Variables
        self.add_input('rotor_torque', 0.0, units='N*m', desc = 'torque from rotor at rated power') #JMF do we want this default?
        self.add_input('gearbox_mass_coeff', 113., desc= 'A in the gearbox mass equation: A*rotor_torque^exp')
        self.add_input('gearbox_mass_exp', 0.71, desc= 'exp in the gearbox mass equation: A*rotor_torque^exp')
        
        # Outputs
        self.add_output('gearbox_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        rotor_torque = inputs['rotor_torque']
        gearbox_mass_coeff = inputs['gearbox_mass_coeff']
        gearbox_mass_exp = inputs['gearbox_mass_exp']
        
        # calculate the gearbox mass
        outputs['gearbox_mass'] = gearbox_mass_coeff * (rotor_torque/1000.0)**gearbox_mass_exp

# --------------------------------------------------------------------
class HighSpeedSideMass(ExplicitComponent):

    def setup(self):
      

        # Variables
        self.add_input('machine_rating', 0.0, units='kW', desc='machine rating')
        self.add_input('hss_mass_coeff', 0.19894, desc= 'NREL CSM hss equation; removing intercept since it is negligible')
        
        # Outputs
        self.add_output('hss_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        machine_rating = inputs['machine_rating']
        hss_mass_coeff = inputs['hss_mass_coeff']
        
        # TODO: this is in DriveSE; replace this with code in DriveSE and have DriveSE use this code??
        outputs['hss_mass'] = hss_mass_coeff * machine_rating

# --------------------------------------------------------------------
class GeneratorMass(ExplicitComponent):

    def setup(self):
      
        
  
        # Variables
        self.add_input('machine_rating', 0.0, units='kW', desc='machine rating')
        self.add_input('generator_mass_coeff', 2300., desc= 'A in the generator mass equation: A*rated_power + B') #default from ppt
        self.add_input('generator_mass_intercept', 3400., desc= 'B in the generator mass equation: A*rated_power + B') #default from ppt
        
        # Outputs
        self.add_output('generator_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):

        machine_rating = inputs['machine_rating']
        generator_mass_coeff = inputs['generator_mass_coeff']
        generator_mass_intercept = inputs['generator_mass_intercept']
    
        # calculate the generator mass
        outputs['generator_mass'] = generator_mass_coeff * machine_rating/1000. + generator_mass_intercept

# --------------------------------------------------------------------
class BedplateMass(ExplicitComponent):

    def setup(self):


        # Variables
        self.add_input('rotor_diameter', 0.0, units='m', desc= 'rotor diameter of the machine')
        self.add_input('bedplate_mass_exp', 2.2, desc= 'exp in the bedplate mass equation: rotor_diameter^exp')
        
        # Outputs
        self.add_output('bedplate_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        rotor_diameter = inputs['rotor_diameter']
        bedplate_mass_exp = inputs['bedplate_mass_exp']
        
        # calculate the bedplate mass
        outputs['bedplate_mass'] = rotor_diameter**bedplate_mass_exp

# --------------------------------------------------------------------
class YawSystemMass(ExplicitComponent):
  
    def setup(self):


        # Variables
        self.add_input('rotor_diameter', 0.0, units='m', desc= 'rotor diameter of the machine')
        self.add_input('yaw_mass_coeff', 0.0009, desc= 'A in the yaw mass equation: A*rotor_diameter^exp') #NREL CSM
        self.add_input('yaw_mass_exp', 3.314, desc= 'exp in the yaw mass equation: A*rotor_diameter^exp') #NREL CSM
        
        # Outputs
        self.add_output('yaw_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
      
        rotor_diameter = inputs['rotor_diameter']
        yaw_mass_coeff = inputs['yaw_mass_coeff']
        yaw_mass_exp = inputs['yaw_mass_exp']
    
        # calculate yaw system mass #TODO - 50% adder for non-bearing mass
        outputs['yaw_mass'] = 1.5 * (yaw_mass_coeff * rotor_diameter ** yaw_mass_exp) #JMF do we really want to expose all these?

#TODO: no variable speed mass; ignore for now

# --------------------------------------------------------------------
class HydraulicCoolingMass(ExplicitComponent):
    
    def setup(self):
    
        
        # Variables
        self.add_input('machine_rating', 0.0, units='kW', desc='machine rating')
        self.add_input('hvac_mass_coeff', 0.08, desc= 'hvac linear coeff') #NREL CSM
        
        # Outputs
        self.add_output('hvac_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        machine_rating = inputs['machine_rating']
        hvac_mass_coeff = inputs['hvac_mass_coeff']
        
        # calculate hvac system mass
        outputs['hvac_mass'] = hvac_mass_coeff * machine_rating

# --------------------------------------------------------------------
class NacelleCoverMass(ExplicitComponent):

    def setup(self):
    
    
        # Variables
        self.add_input('machine_rating', 0.0, units='kW', desc='machine rating')
        self.add_input('cover_mass_coeff', 1.2817, desc= 'A in the spinner mass equation: A*rotor_diameter + B')
        self.add_input('cover_mass_intercept', 428.19, desc= 'B in the spinner mass equation: A*rotor_diameter + B')
        
        # Outputs
        self.add_output('cover_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        machine_rating = inputs['machine_rating']
        cover_mass_coeff = inputs['cover_mass_coeff']
        cover_mass_intercept = inputs['cover_mass_intercept']
        
        # calculate nacelle cover mass
        outputs['cover_mass'] = cover_mass_coeff * machine_rating + cover_mass_intercept

# TODO: ignoring controls and electronics mass for now

# --------------------------------------------------------------------
class OtherMainframeMass(ExplicitComponent):
    # nacelle platforms, service crane, base hardware
    
    def setup(self):
    
        
        # Variables
        self.add_input('bedplate_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('platforms_mass_coeff', 0.125, desc='nacelle platforms mass coeff as a function of bedplate mass [kg/kg]') #default from old CSM
        self.add_discrete_input('crane', False, desc='flag for presence of onboard crane')
        self.add_input('crane_weight', 3000., desc='weight of onboard crane')
        #TODO: there is no base hardware mass model in the old model. Cost is not dependent on mass.
        
        # Outputs
        self.add_output('other_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        
        bedplate_mass = inputs['bedplate_mass']
        platforms_mass_coeff = inputs['platforms_mass_coeff']
        crane = discrete_inputs['crane']
        crane_weight = inputs['crane_weight']
        
        # calculate nacelle cover mass           
        platforms_mass = platforms_mass_coeff * bedplate_mass

        # --- crane ---        
        if (crane):
            crane_mass =  crane_weight
        else:
            crane_mass = 0.  
        
        outputs['other_mass'] = platforms_mass + crane_mass

# --------------------------------------------------------------------
class TransformerMass(ExplicitComponent):

    def setup(self):
    
    
        # Variables
        self.add_input('machine_rating', 0.0, units='kW', desc='machine rating')
        self.add_input('transformer_mass_coeff', 1915., desc= 'A in the transformer mass equation: A*rated_power + B') #default from ppt
        self.add_input('transformer_mass_intercept', 1910., desc= 'B in the transformer mass equation: A*rated_power + B') #default from ppt
        
        # Outputs
        self.add_output('transformer_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        machine_rating = inputs['machine_rating']
        transformer_mass_coeff = inputs['transformer_mass_coeff']
        transformer_mass_intercept = inputs['transformer_mass_intercept']
        
        # calculate the transformer mass
        outputs['transformer_mass'] = transformer_mass_coeff * machine_rating/1000. + transformer_mass_intercept

# --------------------------------------------------------------------
class TowerMass(ExplicitComponent):
  
    def setup(self):


        # Variables
        self.add_input('hub_height', 0.0, desc= 'hub height of wind turbine above ground / sea level')
        self.add_input('tower_mass_coeff', 19.828, desc= 'A in the tower mass equation: A*hub_height^B') #default from ppt
        self.add_input('tower_mass_exp', 2.0282, desc= 'B in the tower mass equation: A*hub_height^B') #default from ppt
        
        # Outputs
        self.add_output('tower_mass', 0.0, units='kg', desc='component mass [kg]')
    
    def compute(self, inputs, outputs):
        
        hub_height = inputs['hub_height']
        tower_mass_coeff = inputs['tower_mass_coeff']
        tower_mass_exp = inputs['tower_mass_exp']
        
        # calculate the tower mass
        outputs['tower_mass'] = tower_mass_coeff * hub_height ** tower_mass_exp
 

# Turbine mass adder
class turbine_mass_adder(ExplicitComponent):
    
    def setup(self):
        
    
        # Inputs
        # rotor
        self.add_input('blade_mass', 0.0, units='kg', desc= 'component mass [kg]')
        self.add_input('hub_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('pitch_system_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('spinner_mass', 0.0, units='kg', desc='component mass [kg]')
        # nacelle
        self.add_input('lss_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('main_bearing_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('gearbox_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('hss_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('generator_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('bedplate_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('yaw_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('hvac_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('cover_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('other_mass', 0.0, units='kg', desc='component mass [kg]')
        self.add_input('transformer_mass', 0.0, units='kg', desc='component mass [kg]')
        # tower
        self.add_input('tower_mass', 0.0, units='kg', desc='component mass [kg]')
    
        # Parameters
        self.add_discrete_input('blade_number', 3, desc = 'number of rotor blades')
        self.add_discrete_input('bearing_number', 2, desc = 'number of main bearings')
    
        # Outputs
        self.add_output('hub_system_mass', 0.0, units='kg', desc='hub system mass')
        self.add_output('rotor_mass', 0.0, units='kg', desc='hub system mass')
        self.add_output('nacelle_mass', 0.0, units='kg', desc='nacelle mass')
        self.add_output('turbine_mass', 0.0, units='kg', desc='turbine mass')
    
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        blade_mass = inputs['blade_mass']
        hub_mass = inputs['hub_mass']
        pitch_system_mass = inputs['pitch_system_mass']
        spinner_mass = inputs['spinner_mass']
        lss_mass = inputs['lss_mass']
        main_bearing_mass = inputs['main_bearing_mass']
        gearbox_mass = inputs['gearbox_mass']
        hss_mass = inputs['hss_mass']
        generator_mass = inputs['generator_mass']
        bedplate_mass = inputs['bedplate_mass']
        yaw_mass = inputs['yaw_mass']
        hvac_mass = inputs['hvac_mass']
        cover_mass = inputs['cover_mass']
        other_mass = inputs['other_mass']
        transformer_mass = inputs['transformer_mass']
        tower_mass = inputs['tower_mass']
        blade_number = discrete_inputs['blade_number']
        bearing_number = discrete_inputs['bearing_number']
        
        
        outputs['hub_system_mass'] = hub_mass + pitch_system_mass + spinner_mass
        outputs['rotor_mass'] = blade_mass * blade_number + outputs['hub_system_mass']
        outputs['nacelle_mass'] = lss_mass + bearing_number * main_bearing_mass + \
                            gearbox_mass + hss_mass + generator_mass + \
                            bedplate_mass + yaw_mass + hvac_mass + \
                            cover_mass + other_mass + transformer_mass
        outputs['turbine_mass'] = outputs['rotor_mass'] + outputs['nacelle_mass'] + tower_mass

# --------------------------------------------------------------------

class nrel_csm_mass_2015(Group):
    
    def setup(self):

        sharedIndeps = IndepVarComp()
        sharedIndeps.add_output('machine_rating',     units='kW', val=0.0)
        sharedIndeps.add_output('rotor_diameter',         units='m', val=0.0)
        sharedIndeps.add_discrete_output('blade_number',  val=0)
        sharedIndeps.add_discrete_output('crane', val=False)
        self.add_subsystem('sharedIndeps', sharedIndeps, promotes=['*'])

        self.add_subsystem('blade',BladeMass(), promotes=['*'])
        self.add_subsystem('hub',HubMass(), promotes=['*'])
        self.add_subsystem('pitch',PitchSystemMass(), promotes=['*'])
        self.add_subsystem('spinner',SpinnerMass(), promotes=['*'])
        self.add_subsystem('lss',LowSpeedShaftMass(), promotes=['*'])
        self.add_subsystem('bearing',BearingMass(), promotes=['*'])
        self.add_subsystem('torque',RotorTorque(), promotes=['*'])
        self.add_subsystem('gearbox',GearboxMass(), promotes=['*'])
        self.add_subsystem('hss',HighSpeedSideMass(), promotes=['*'])
        self.add_subsystem('generator',GeneratorMass(), promotes=['*'])
        self.add_subsystem('bedplate',BedplateMass(), promotes=['*'])
        self.add_subsystem('yaw',YawSystemMass(), promotes=['*'])
        self.add_subsystem('hvac',HydraulicCoolingMass(), promotes=['*'])
        self.add_subsystem('cover',NacelleCoverMass(), promotes=['*'])
        self.add_subsystem('other',OtherMainframeMass(), promotes=['*'])
        self.add_subsystem('transformer',TransformerMass(), promotes=['*'])
        self.add_subsystem('tower',TowerMass(), promotes=['*'])
        self.add_subsystem('turbine',turbine_mass_adder(), promotes=['*'])
       

class nrel_csm_2015(Group):
    
    def setup(self):
        self.add_subsystem('nrel_csm_mass', nrel_csm_mass_2015(), promotes=['*'])
        self.add_subsystem('turbine_costs', Turbine_CostsSE_2015(verbosity=False, topLevelFlag=False), promotes=['*'])

#-----------------------------------------------------------------

def mass_example():

    # simple test of module
    trb = nrel_csm_mass_2015()
    prob = Problem(trb)
    prob.setup()
    
    prob['rotor_diameter'] = 126.0
    prob['turbine_class'] = 1
    prob['blade_has_carbon'] = False
    prob['blade_number'] = 3    
    prob['machine_rating'] = 5000.0
    prob['hub_height'] = 90.0
    prob['bearing_number'] = 2
    prob['crane'] = True
    prob['max_tip_speed'] = 80.0
    prob['max_efficiency'] = 0.90

    prob.run_model()
   
    print("The MASS results for the NREL 5 MW Reference Turbine in an offshore 20 m water depth location are:")
    #print "Overall turbine mass is {0:.2f} kg".format(trb.turbine.params['turbine_mass'])
    for io in trb._outputs:
        print(io, str(trb._outputs[io]))

def cost_example():

    # simple test of module
    trb = nrel_csm_2015()
    prob = Problem(trb)
    prob.setup()

    # simple test of module
    prob['rotor_diameter'] = 126.0
    prob['turbine_class'] = 1
    prob['blade_has_carbon'] = False
    prob['blade_number'] = 3    
    prob['machine_rating'] = 5000.0
    prob['hub_height'] = 90.0
    prob['bearing_number'] = 2
    prob['crane'] = True
    prob['max_tip_speed'] = 80.0
    prob['max_efficiency'] = 0.90

    prob.run_model()

    print("The COST results for the NREL 5 MW Reference Turbine:")
    for io in trb._outputs:
        print(io, str(trb._outputs[io]))

if __name__ == "__main__":
    mass_example()
    cost_example()



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.pBeam._pBEAM as pb

class TestPyBeam(unittest.TestCase):
    
    def testCantileverDeflection(self):
        # Test data from "Finite Element Structural Analysis", Yang, pg. 145
        E = 2.0
        I = 3.0
        L = 4.0
        p0 = 5.0

        nodes = 2

        Px = np.array([-p0, 0.0])
        Py = Pz = np.zeros(nodes)
        loads = pb.Loads(nodes, Px, Py, Pz)

        tip = pb.TipData()

        base = pb.BaseData(np.ones(6), 1.0)

        z = np.array([0.0, L])
        EIx = EIy = E*I*np.ones(nodes)
        EA = E*np.ones(nodes)
        GJ = rhoA = rhoJ = np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        beam = pb.Beam(sec, loads, tip, base)

        (dx, dy, dz, dtheta_x, dtheta_y, dtheta_z) = beam.displacement()

        self.assertAlmostEqual(dx[0], 0.0, 8)
        self.assertAlmostEqual(dx[1], -p0 * L**3.0 / E / I * L / 30.0, 8)
        self.assertAlmostEqual(dtheta_y[0], 0.0, 8)
        self.assertAlmostEqual(dtheta_y[1],-p0 * L**3.0 / E / I * 1 / 24.0, 8)

    def testTaperedDeflections(self):
        # Test data from "Finite Element Structural Analysis", Yang, pg. 180
        E = 2.0
        I = 3.0
        L = 4.0
        P = 5.0

        nodes = 3

        Px = Py = Pz = np.zeros(nodes)
        loads = pb.Loads(nodes, Px, Py, Pz)

        tip = pb.TipData(0.0, np.zeros(3), np.zeros(6), [-P, 0.0, 0.0], np.zeros(3))

        base = pb.BaseData(np.ones(6), 1.0)

        z = np.array([0.0, 0.5*L, L])
        EIx = EIy = E*I*np.array([9.0, 5.0, 1.0])
        EA = E*np.ones(nodes)
        GJ = rhoA = rhoJ = np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        beam = pb.Beam(sec, loads, tip, base)

        (dx, dy, dz, dtheta_x, dtheta_y, dtheta_z) = beam.displacement()

        tol = 1e-8
        tol_pct_1 = 0.17;
        tol_pct_2 = 0.77;
        self.assertAlmostEqual(dx[0], 0.0, 8)
        self.assertAlmostEqual(dx[-1], -0.051166*P*L**3/E/I, delta=tol_pct_1)
        self.assertAlmostEqual(dtheta_y[0], 0.0, 8)
        self.assertAlmostEqual(dtheta_y[-1], -0.090668*P*L**2/E/I, delta=tol_pct_2)

    def testFreqFree_FreeBeam_n1(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0

        n = 1
        nodes = n+1
        
        Px = Py = Pz = np.zeros(nodes)
        loads = pb.Loads(nodes, Px, Py, Pz)

        tip = pb.TipData()

        base = pb.BaseData()
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        beam = pb.Beam(sec, loads, tip, base)

        nFreq = 100
        freq = beam.naturalFrequencies(nFreq)
        
        m = rho * A
        alpha = m * (n*L)**4.0 / (840.0 * E * I)

        tol_pct = 5e-6 * 100
        expect = np.sqrt(0.85714 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[1], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[2], expect, delta=tol_pct)
        expect = np.sqrt(10.0 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[4], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[5], expect, delta=tol_pct)


    def testFreqFree_FreeBeam_n2(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0

        n = 2
        nodes = n+1
        
        Px = Py = Pz = np.zeros(nodes)
        loads = pb.Loads(nodes, Px, Py, Pz)

        tip = pb.TipData()

        base = pb.BaseData()
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        beam = pb.Beam(sec, loads, tip, base)

        nFreq = 100
        freq = beam.naturalFrequencies(nFreq)
        
        m = rho * A
        alpha = m * (n*L)**4.0 / (840.0 * E * I)

        tol_pct = 6e-6 * 100
        expect = np.sqrt(0.59858 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[1], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[2], expect, delta=tol_pct)
        expect = np.sqrt(5.8629 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[5], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[6], expect, delta=tol_pct)
        expect = np.sqrt(36.659 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[8], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[9], expect, delta=tol_pct)
        expect = np.sqrt(93.566 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[10], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[11], expect, delta=tol_pct)
        

    def testFreqFree_FreeBeam_n3(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0

        n = 3
        nodes = n+1
        
        Px = Py = Pz = np.zeros(nodes)
        loads = pb.Loads(nodes, Px, Py, Pz)

        tip = pb.TipData()

        base = pb.BaseData()
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        beam = pb.Beam(sec, loads, tip, base)

        nFreq = 100
        freq = beam.naturalFrequencies(nFreq)
        
        m = rho * A
        alpha = m * (n*L)**4.0 / (840.0 * E * I)

        tol_pct = 6e-6 * 100
        expect = np.sqrt(0.59919 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[1], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[2], expect, delta=tol_pct)
        expect = np.sqrt(4.5750 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[5], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[6], expect, delta=tol_pct)
        expect = np.sqrt(22.010 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[8], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[9], expect, delta=tol_pct)
        expect = np.sqrt(70.920 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[11], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[12], expect, delta=tol_pct)
        expect = np.sqrt(265.91 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[14], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[15], expect, delta=tol_pct)
        expect = np.sqrt(402.40 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[16], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[17], expect, delta=tol_pct)

    def testBucklingEuler(self):
        # unit test data from Euler's buckling formula for a clamped/free beam

        E = 2.0
        I = 3.0
        L = 4.0

        n = 3
        nodes = n+1
        
        Px = Py = Pz = np.zeros(nodes)
        loads = pb.Loads(nodes, Px, Py, Pz)

        tip = pb.TipData()

        base = pb.BaseData(np.ones(6), 1.0)
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = rhoA = np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        beam = pb.Beam(sec, loads, tip, base)

        (Pcr_x, Pcr_y) = beam.criticalBucklingLoads()

        Ltotal = n*L
        tol_pct = 0.011
        expect = E * I * (0.5*np.pi/Ltotal)**2.0
        self.assertAlmostEqual(Pcr_x, expect, 2)
        self.assertAlmostEqual(Pcr_y, expect, 2)

    def testShearBendingSimple(self):
        # Test data from "Mechanical of Materials", Gere, 6th ed., pg. 273
        # cantilevered beam with linear distributed load

        L = 10.0
        q0 = 3.0

        n = 1
        nodes = n+1

        tip = pb.TipData()

        base = pb.BaseData(np.ones(6), 1.0)
        
        z = np.arange(nodes, dtype=np.float64) * (L / n)
        EIx = EIy = EA = GJ = rhoJ = rhoA = np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        Px = q0*(1 - z/L)
        Py = Pz = np.zeros(nodes)
        loads = pb.Loads(nodes, Px, Py, Pz)

        beam = pb.Beam(sec, loads, tip, base)

        (Vx, Vy, Fz, Mx, My, Tz) = beam.shearAndBending()

        tol_pct = 1e-8
        Vx_expect = np.polyval([q0*L/2.0, -q0*L, q0*L/2.0], [0.0, 1.0])
        My_expect = np.polyval([-q0*L*L/6.0, 3.0*q0*L*L/6.0, -3.0*q0*L*L/6.0, q0*L*L/6.0], [0.0, 1.0])
        npt.assert_almost_equal(Vx, Vx_expect)
        npt.assert_almost_equal(My, My_expect)

    def testShearBendingSimplePt(self):
        # Test data from "Mechanical of Materials", Gere, 6th ed., pg. 288
        # cantilevered beam with two point loads

        L = 10.0
        P1 = 2.0
        P2 = 3.0

        n = 3
        nodes = n+1

        tip = pb.TipData()

        base = pb.BaseData(np.ones(6), 1.0)
        
        z = np.arange(nodes, dtype=np.float64) * (L / n)
        EIx = EIy = EA = GJ = rhoJ = rhoA = np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        Px = Py = Pz = Fy_pt = Fz_pt = Mx_pt = My_pt = Mz_pt = np.zeros(nodes)
        Fx_pt = np.zeros(nodes)
        Fx_pt[1] = -P2
        Fx_pt[3] = -P1
        loads = pb.Loads(nodes, Px, Py, Pz, Fx_pt, Fy_pt, Fz_pt, Mx_pt, My_pt, Mz_pt)

        beam = pb.Beam(sec, loads, tip, base)

        (Vx, Vy, Fz, Mx, My, Tz) = beam.shearAndBending()

        tol_pct = 1e-8;
        Vx_expect = np.polyval([0.0, 0.0, -P1-P2], [0.0])
        self.assertAlmostEqual(Vx[0], Vx_expect, 8)

        Vx_expect = np.polyval([0.0, 0.0, -P1], [0.0])
        self.assertAlmostEqual(Vx[1], Vx_expect, 8)
        self.assertAlmostEqual(Vx[2], Vx_expect, 8)

        b = L/3.0
        a = 2.0/3.0*L
        My_expect = np.polyval([0.0, 0.0, -P1*a + P1*L + P2*b, -P1*L - P2*b], [0.0])
        self.assertAlmostEqual(My[0], My_expect, 8)

        My_expect = np.polyval([0.0, 0.0, -0.5*P1*a + P1*a, -P1*a], [0.0])
        self.assertAlmostEqual(My[1], My_expect, 8)

        My_expect = np.polyval([0.0, 0.0, 0.5*P1*a, -0.5*P1*a], [0.0])
        self.assertAlmostEqual(My[2], My_expect, 8)

    def testOtherCalls(self):

        E = 2.0
        I = 3.0
        L = 4.0
        q0 = 5.0
        
        n = 3
        nodes = n+1
        
        tip = pb.TipData()

        base = pb.BaseData(np.ones(6), 1.0)
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = rhoA = np.ones(nodes)
        sec = pb.SectionData(nodes, z, EA, EIx, EIy, GJ, rhoA, rhoJ)

        Pz = q0*(1 - z/L)
        Py = Px = np.zeros(nodes)
        loads = pb.Loads(nodes, Px, Py, Pz)

        beam = pb.Beam(sec, loads, tip, base)

        badlist = [float('inf'), -float('inf'), float('nan'), 0.0, np.inf, -np.inf, np.nan]
        self.assertNotIn(beam.mass(), badlist)
        self.assertNotIn(beam.outOfPlaneMomentOfInertia(), badlist)

        npts = 10
        xv = yv = np.zeros(npts)
        zv = np.linspace(z[0], z[-1], npts)
        self.assertNotIn( beam.axialStrain(npts, xv, yv, zv).sum(), badlist)

    def testCurveFEM_FixedBeam_n1(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0
        hubR = 0.0
        omegaRPM = 0.0

        n = 1
        nodes = n+1
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        theta = precurv = presweep = np.zeros(nodes) #np.linspace(0.0, 3.0, nodes)

        mycurve = pb.CurveFEM(omegaRPM, theta, z, precurv, presweep, rhoA, True)
        freq = mycurve.frequencies(EA, EIx, EIy, GJ, rhoJ, nodes)[0]

        #truth = _curvefem.frequencies(omegaRPM, L, hubR, z/L, theta, rhoA, EIx, EIy, GJ, EA, rhoJ, precurv, presweep)
        #print "n1", truth.tolist()
        truth = np.array([0.012572866969753228, 0.012591740426479996, 0.015715395588976708, 0.015715395588976708, 0.02528529867098764, 0.02528529867098764, 0.06883002554331129, 0.06893334807998643, 0.09116741813672731, 0.09116741813672731, 0.1548391375450802, 0.1548391375450802])
        npt.assert_almost_equal(freq, truth)

    def testCurveFEM_FixedBeam_n2(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0
        hubR = 0.0
        omegaRPM = 0.0

        n = 2
        nodes = n+1
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        theta = precurv = presweep = np.zeros(nodes) #np.linspace(0.0, 3.0, nodes)

        mycurve = pb.CurveFEM(omegaRPM, theta, z, precurv, presweep, rhoA, True)
        freq = mycurve.frequencies(EA, EIx, EIy, GJ, rhoJ, nodes)[0]

        #truth = _curvefem.frequencies(omegaRPM, L, hubR, z/L, theta, rhoA, EIx, EIy, GJ, EA, rhoJ, precurv, presweep)
        #print "n2", truth.tolist()
        truth = np.array([0.003912151523474228, 0.003912151523474228, 0.005852976699216038, 0.012582302518525731, 0.02044674778626943, 0.02471287957376343, 0.02471287957376343, 0.025285550340850595, 0.025285550340850595, 0.032042058241816634, 0.06888168035399123, 0.0835842095344649, 0.0835842095344649, 0.09116764495898616, 0.09116764495898616, 0.11193550172703298, 0.24259762924185935, 0.24259762924185935])
        npt.assert_almost_equal(freq, truth)
        

    def testCurveFEM_FixedBeam_n3(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0
        hubR = 0.0
        omegaRPM = 0.0

        n = 3
        nodes = n+1
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        theta = precurv = presweep = np.zeros(nodes) #np.linspace(0.0, 3.0, nodes)

        mycurve = pb.CurveFEM(omegaRPM, theta, z, precurv, presweep, rhoA, True)
        freq = mycurve.frequencies(EA, EIx, EIy, GJ, rhoJ, nodes)[0]

        #truth = _curvefem.frequencies(omegaRPM, L, hubR, z/L, theta, rhoA, EIx, EIy, GJ, EA, rhoJ, precurv, presweep)
        #print "n3", truth.tolist()
        truth = np.array([0.0017380701632556908, 0.0017380701632562416, 0.003847212436706601, 0.010926964080897513, 0.010926964080897626, 0.012576854615234626, 0.012587751208210572, 0.02106152327278283, 0.02282612879817582, 0.02528527062288509, 0.02528527062288511, 0.03087567297329799, 0.030875672973298095, 0.06885185586578292, 0.06891150914730364, 0.06953079721389932, 0.06953079721389936, 0.09116748336205782, 0.09116748336205784, 0.12496139758839682, 0.1308572353442806, 0.13085723534428065, 0.2608788373591399, 0.2608788373591401])
        npt.assert_almost_equal(freq, truth)
        

    def testCurveFEM_FixedBeam_n3_withShape(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0
        hubR = 0.0
        omegaRPM = 12.0

        n = 3
        nodes = n+1
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        theta = precurv = presweep = np.linspace(0.0, 3.0, nodes)

        mycurve = pb.CurveFEM(omegaRPM, theta, z, precurv, presweep, rhoA, True)
        freq = mycurve.frequencies(EA, EIx, EIy, GJ, rhoJ, nodes)[0]

        #truth = _curvefem.frequencies(omegaRPM, L, hubR, z/L, theta, rhoA, EIx, EIy, GJ, EA, rhoJ, precurv, presweep)
        #print "n3 odd", truth.tolist()
        truth = np.array([0.00583962278555471, 0.013747011747362698, 0.019857000185537085, 0.022065457582918686, 0.022508587859798306, 0.06494224803437999, 0.07593365714823054, 0.08099501901331865, 0.08100059834010384, 0.11781473080899654, 0.1973084679918305, 0.44555268080407895, 0.48641244353481683, 0.7481252945270574, 0.7730228919707995, 1.0518193329242682, 1.0696357241743277, 1.4148111654258362, 1.4280996628456801, 1.9080030090517186, 1.9178773250559262, 0.0, 0.0, 0.0])
        npt.assert_almost_equal(freq[truth>0.0], truth[truth>0.0], decimal=5)

        
    def testCurveFEM_FreeBeam_n1(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0
        hubR = 0.0
        omegaRPM = 0.0

        n = 1
        nodes = n+1
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        theta = precurv = presweep = np.zeros(nodes) #np.linspace(0.0, 3.0, nodes)

        mycurve = pb.CurveFEM(omegaRPM, theta, z, precurv, presweep, rhoA, False)
        freq = mycurve.frequencies(EA, EIx, EIy, GJ, rhoJ, nodes)[0]

        
        m = rho * A
        alpha = m * (n*L)**4.0 / (840.0 * E * I)

        tol_pct = 5e-6 * 100
        expect = np.sqrt(0.85714 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[1], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[2], expect, delta=tol_pct)
        expect = np.sqrt(10.0 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[4], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[5], expect, delta=tol_pct)


    def testCurveFEM_FreeBeam_n2(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0
        hubR = 0.0
        omegaRPM = 0.0

        n = 2
        nodes = n+1
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        theta = precurv = presweep = np.zeros(nodes) #np.linspace(0.0, 3.0, nodes)

        mycurve = pb.CurveFEM(omegaRPM, theta, z, precurv, presweep, rhoA, False)
        freq = mycurve.frequencies(EA, EIx, EIy, GJ, rhoJ, nodes)[0]
        
        m = rho * A
        alpha = m * (n*L)**4.0 / (840.0 * E * I)

        tol_pct = 6e-6 * 100
        expect = np.sqrt(0.59858 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[1], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[2], expect, delta=tol_pct)
        expect = np.sqrt(5.8629 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[5], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[6], expect, delta=tol_pct)
        expect = np.sqrt(36.659 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[8], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[9], expect, delta=tol_pct)
        expect = np.sqrt(93.566 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[10], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[11], expect, delta=tol_pct)

        

    def testCurveFEM_FreeBeam_n3(self):
        # Test data from "Consistent Mass Matrix for Distributed Mass Systmes", John Archer,
        # Journal of the Structural Division Proceedings of the American Society of Civil Engineers,
        # pg. 168
        E = 2.0
        I = 3.0
        L = 4.0
        A = 5.0
        rho = 6.0
        hubR = 0.0
        omegaRPM = 0.0

        n = 3
        nodes = n+1
        
        z = L * np.arange(nodes)
        EIx = EIy = E*I*np.ones(nodes)
        EA = GJ = rhoJ = np.ones(nodes)
        rhoA = rho*A*np.ones(nodes)
        theta = precurv = presweep = np.zeros(nodes) #np.linspace(0.0, 3.0, nodes)

        mycurve = pb.CurveFEM(omegaRPM, theta, z, precurv, presweep, rhoA, False)
        freq = mycurve.frequencies(EA, EIx, EIy, GJ, rhoJ, nodes)[0]
        
        m = rho * A
        alpha = m * (n*L)**4.0 / (840.0 * E * I)

        tol_pct = 6e-6 * 100
        expect = np.sqrt(0.59919 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[1], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[2], expect, delta=tol_pct)
        expect = np.sqrt(4.5750 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[5], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[6], expect, delta=tol_pct)
        expect = np.sqrt(22.010 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[8], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[9], expect, delta=tol_pct)
        expect = np.sqrt(70.920 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[11], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[12], expect, delta=tol_pct)
        expect = np.sqrt(265.91 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[14], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[15], expect, delta=tol_pct)
        expect = np.sqrt(402.40 / alpha) / (2*np.pi)
        self.assertAlmostEqual(freq[16], expect, delta=tol_pct)
        self.assertAlmostEqual(freq[17], expect, delta=tol_pct)

        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestPyBeam))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())



from openmdao.api import ExplicitComponent
import numpy as np
import os
import sys
from wisdem.pymap import pyMAP

from wisdem.commonse import gravity, Enum
from wisdem.commonse.utilities import assembleI, unassembleI

Anchor    = Enum('DRAGEMBEDMENT SUCTIONPILE')
NLINES_MAX = 15
NPTS_PLOT = 20

class MapMooring(ExplicitComponent):
    """
    OpenMDAO Component class for mooring system attached to sub-structure of floating offshore wind turbines.
    Should be tightly coupled with Spar class for full system representation.
    """

    def setup(self):
    
        # Variables local to the class and not OpenMDAO
        self.min_break_load      = None
        self.wet_mass_per_length = None
        self.axial_stiffness     = None
        self.area                = None
        self.cost_per_length     = None
        self.finput              = None
        self.tlpFlag             = False

        # Environment
        self.add_input('water_density', val=0.0, units='kg/m**3', desc='density of water')
        self.add_input('water_depth', val=0.0, units='m', desc='water depth')

        # Material properties

        # Inputs from SparGeometry
        self.add_input('fairlead_radius', val=0.0, units='m', desc='Outer spar radius at fairlead depth (point of mooring attachment)')
        
        # Design variables
        self.add_input('fairlead', val=0.0, units='m', desc='Depth below water for mooring line attachment')
        self.add_input('mooring_line_length', val=0.0, units='m',desc='Unstretched total mooring line length')
        self.add_input('anchor_radius', val=0.0, units='m', desc='radius from center of spar to mooring anchor point')
        self.add_input('mooring_diameter', val=0.0, units='m',desc='diameter of mooring line')

        # User inputs (could be design variables)
        self.add_input('number_of_mooring_connections', val=3, desc='number of mooring connections on vessel')
        self.add_input('mooring_lines_per_connection', val=1, desc='number of mooring lines per connection')
        self.add_discrete_input('mooring_type', val='CHAIN', desc='chain, nylon, polyester, fiber, or iwrc')
        self.add_discrete_input('anchor_type', val='DRAGEMBEDMENT', desc='SUCTIONPILE or DRAGEMBEDMENT')
        self.add_input('max_offset', val=0.0, units='m',desc='X offsets in discretization')
        self.add_input('operational_heel', val=0.0, units='deg',desc='Maximum angle of heel allowable during operation')
        self.add_input('max_survival_heel', val=0.0, units='deg', desc='max heel angle for turbine survival')
        self.add_input('gamma_f', val=0.0, desc='Safety factor for mooring line tension')

        # Cost rates
        self.add_input('mooring_cost_factor', val=0.0, desc='miscellaneous cost factor in percent')

        # Outputs
        self.add_output('number_of_mooring_lines', val=0, desc='total number of mooring lines')
        self.add_output('mooring_mass', val=0.0, units='kg',desc='total mass of mooring')
        self.add_output('mooring_moments_of_inertia', val=np.zeros(6), units='kg*m**2', desc='mass moment of inertia of mooring system about fairlead-centerline point [xx yy zz xy xz yz]')
        self.add_output('mooring_cost', val=0.0, units='USD',desc='total cost for anchor + legs + miscellaneous costs')
        self.add_output('mooring_stiffness', val=np.zeros((6,6)), units='N/m', desc='Linearized stiffness matrix of mooring system at neutral (no offset) conditions.')
        self.add_output('anchor_cost', val=0.0, units='USD',desc='total cost for anchor')
        self.add_output('mooring_neutral_load', val=np.zeros((NLINES_MAX,3)), units='N',desc='mooring vertical load in all mooring lines')
        self.add_output('max_offset_restoring_force', val=0.0, units='N',desc='sum of forces in x direction after max offset')
        self.add_output('operational_heel_restoring_force', val=np.zeros((NLINES_MAX,3)), units='N',desc='forces for all mooring lines after operational heel')
        self.add_output('mooring_plot_matrix', val=np.zeros((NLINES_MAX, NPTS_PLOT, 3)), units='m', desc='data matrix for plotting') 

        # Output constriants
        self.add_output('axial_unity', val=0.0, units='m',desc='range of damaged mooring')
        self.add_output('mooring_length_max', val=0.0, desc='mooring line length ratio to nodal distance')
        
        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        """Sets mooring line properties then writes MAP input file and executes MAP.
        
        INPUTS:
        ----------
        inputs   : dictionary of input parameters
        outputs : dictionary of output parameters
        
        OUTPUTS  : none (all unknown dictionary values set)
        """

        # Set characteristics based on regressions / empirical data
        self.set_properties(inputs, discrete_inputs)

        # Set geometry profile
        self.set_geometry(inputs, outputs)

        # Write MAP input file and analyze the system at every angle
        self.runMAP(inputs, discrete_inputs, outputs)

        # Compute costs for the system
        self.compute_cost(inputs, discrete_inputs, outputs)


    def set_properties(self, inputs, discrete_inputs):
        """Sets mooring line properties: Minimum Breaking Load, Mass per Length, Axial Stiffness, Cross-Sectional Area, Cost-per-Length.
        
        INPUTS:
        ----------
        inputs   : dictionary of input parameters
        
        OUTPUTS  : Parameters are class variables and are set internally

        References:
        https://daim.idi.ntnu.no/masteroppgaver/015/15116/masteroppgave.pdf
        http://offshoremechanics.asmedigitalcollection.asme.org/article.aspx?articleid=2543338
        https://www.orcina.com/SoftwareProducts/OrcaFlex/Documentation/Help/Content/html/
        Chain.htm
        Chain,AxialandBendingStiffness.htm
        Chain,MechanicalProperties.htm
        RopeWire.htm
        RopeWire,MinimumBreakingLoads.htm
        RopeWire,Massperunitlength.htm
        RopeWire,AxialandBendingStiffness.htm
        """
        
        # Unpack variables
        Dmooring       = inputs['mooring_diameter']
        lineType       = discrete_inputs['mooring_type'].upper()

        # Set parameters based on regressions for different mooring line type
        Dmooring2 = Dmooring**2

        # TODO: Costs per unit length are not synced with new input sources
        if lineType == 'CHAIN':
            self.min_break_load      = 2.74e7  * Dmooring2 * (44.0 - 80.0*Dmooring)
            # Use a linear fit to the other fit becuase it is poorly conditioned for optimization
            #self.min_break_load      = 1e3*np.maximum(1.0, -5445.2957034820683+176972.68498888266*Dmooring)
            self.wet_mass_per_length = 19.9e3  * Dmooring2 # From Orca, 7983.34117 OC3 definiton doc
            self.axial_stiffness     = 8.54e10 * Dmooring2 # From Orca, 4.74374e10 OC3 definiton doc, 
            self.area                = 2.0 * 0.25 * np.pi * Dmooring2
            self.cost_per_length     = 3.415e4  * Dmooring2 #0.58*1e-3*self.min_break_load/gravity - 87.6

        elif lineType == 'NYLON':
            self.min_break_load      = 139357e3 * Dmooring2
            self.wet_mass_per_length = 0.6476e3 * Dmooring2
            self.axial_stiffness     = 1.18e8   * Dmooring2
            self.area                = 0.25 * np.pi * Dmooring2
            self.cost_per_length     = 3.415e4  * Dmooring2 #0.42059603*1e-3*self.min_break_load/gravity + 109.5

        elif lineType == 'POLYESTER':
            self.min_break_load      = 170466e3 * Dmooring2
            self.wet_mass_per_length = 0.7978e3 * Dmooring2
            self.axial_stiffness     = 1.09e9   * Dmooring2
            self.area                = 0.25 * np.pi * Dmooring2
            self.cost_per_length     = 3.415e4  * Dmooring2 #0.42059603*1e-3*self.min_break_load/gravity + 109.5

        elif lineType == 'FIBER': # Wire rope with fiber rope
            self.min_break_load      = 584175e3 * Dmooring2
            self.wet_mass_per_length = 3.6109e3 * Dmooring2
            self.axial_stiffness     = 3.67e10  * Dmooring2
            self.area                = 0.455 * 0.25 * np.pi * Dmooring2
            self.cost_per_length     = 2.0 * 6.32e4  * Dmooring2 #0.53676471*1e-3*self.min_break_load/gravity

        elif lineType == 'IWRC': # Wire rope with steel core
            self.min_break_load      = 633358e3 * Dmooring2
            self.wet_mass_per_length = 3.9897e3 * Dmooring2
            self.axial_stiffness     = 4.04e10  * Dmooring2
            self.area                = 0.455 * 0.25 * np.pi * Dmooring2
            self.cost_per_length     = 6.32e4  * Dmooring2 #0.33*1e-3*self.min_break_load/gravity + 139.5

        else:
            raise ValueError('Available line types are: chain nylon polyester fiber iwrc')

            
    def set_geometry(self, inputs, outputs):
        # Unpack variables
        fairleadDepth = inputs['fairlead']
        R_fairlead    = inputs['fairlead_radius']
        R_anchor      = inputs['anchor_radius']
        waterDepth    = inputs['water_depth']
        L_mooring     = inputs['mooring_line_length']
        max_heel      = inputs['max_survival_heel']
        gamma         = inputs['gamma_f']
        
        if L_mooring > (waterDepth - fairleadDepth):
            self.tlpFlag = False
            
            # Create constraint that line isn't too long that there is no catenary hang
            outputs['mooring_length_max'] = L_mooring / (0.95 * (R_anchor + waterDepth - fairleadDepth) )
        else:
            self.tlpFlag = True
            # Create constraint that we don't lose line tension
            outputs['mooring_length_max'] = L_mooring / ( (waterDepth - fairleadDepth - gamma*R_fairlead*np.sin(np.deg2rad(max_heel))) )

    
    def write_line_dictionary(self, inputs, discrete_inputs, cable_sea_friction_coefficient=0.65):
        """Writes LINE DICTIONARY section of input.map file
        
        INPUTS:
        ----------
        inputs   : dictionary of input parameters
        cable_sea_friction_coefficient : coefficient of friction with sea floor (defaults to 0.65)
        
        OUTPUTS  : none
        """
        # Unpack variables
        rhoWater = inputs['water_density']
        lineType = discrete_inputs['mooring_type'].lower()
        Dmooring = inputs['mooring_diameter']
        
        self.finput.append('---------------------- LINE DICTIONARY ---------------------------------------')
        self.finput.append('LineType  Diam      MassDenInAir   EA            CB   CIntDamp  Ca   Cdn    Cdt')
        self.finput.append('(-)       (m)       (kg/m)        (N)           (-)   (Pa-s)    (-)  (-)    (-)')
        self.finput.append('%s   %.5f   %.5f   %.5f   %.5f   1.0E8   0.6   -1.0   0.05' %
                          (lineType, Dmooring, self.wet_mass_per_length, self.axial_stiffness, cable_sea_friction_coefficient) )

        
    def write_node_properties_header(self):
        """Writes NODE PROPERTIES section header of input.map file
        
        INPUTS: none 
        ----------
        
        OUTPUTS  : none
        """
        self.finput.append('---------------------- NODE PROPERTIES ---------------------------------------')
        # Doesn't like some weird character here somewhere
        #self.finput.append('Node  Type       X       Y       Z      M     B     FX      FY      FZ')
        #self.finput.append('(-)   (-)       (m)     (m)     (m)    (kg)  (m^3)  (N)     (N)     (N)')
        self.finput.append('Node Type X     Y    Z   M     V FX FY FZ')
        self.finput.append('(-)  (-) (m)   (m)  (m) (kg) (m^3) (kN) (kN) (kN)')



    def write_node_properties(self, number, node_type, x_pos, y_pos, z_pos,
                              point_mass=0, displaced_volume=0,
                              x_force=None, y_force=None, z_force=None):
        """Writes NODE PROPERTIES data of input.map file.  Nodes are connections between mooring lines and bridles, vessels, and anchors
        
        INPUTS:
        ----------
        number           : The node number listing in the input file
        node_type        : fix / vessel / connect (fix=anchor)
        x_, y_, z_pos    : position of node in coordinate system (separate inputs)
        point_mass       : see MAP reference (defaults to 0)
        displaced_volume : see MAP reference (defaults to 0)
        x_, y_, z_force  : see MAP reference (defaults to None)
        
        OUTPUTS  : none
        """
        # Ensure this connection is something MAP understands
        nodeStr = node_type.lower()
        if not nodeStr in ['fix', 'connect', 'vessel']:
            raise ValueError('%s is not a valid node type for node %d' % (node_type, number))

        # If this is a node between two lines have to specify connection details
        if nodeStr == 'connect':
            try:
                x_force = float(x_force)
                y_force = float(y_force)
                z_force = float(z_force)
            except:
                raise ValueError('%s must have numerical force applied values.' % node_type)

        # Set location strings
        forceStr = '#   #   #' 
        if nodeStr == 'connect':
            forceStr = '%f   %f   %f' % (x_force, y_force, z_force)
            posStr   = '#%f   #%f   #%f   ' % (x_pos, y_pos, z_pos)
        elif nodeStr == 'fix':
            posStr   = '%f   %f   depth   ' % (x_pos, y_pos)
        elif nodeStr == 'vessel':
            posStr   = '%f   %f   %f   ' % (x_pos, y_pos, z_pos)

        # Write the connection line
        line = ('%d   ' % number)
        line += ('%s   ' % node_type)
        line += (posStr)
        line += ('%f   %f   ' % (point_mass, displaced_volume) )
        line += (forceStr)
        self.finput.append(line)

    def write_line_properties(self, inputs, discrete_inputs, line_number=1, anchor_node=2, fairlead_node=1, flags=''):
        """Writes LINE PROPERTIES section of input.map file that connects multiple nodes
        
        INPUTS:
        ----------
        inputs        : dictionary of input parameters- only 'mooring_type' is used
        line_number   : Line ID number (defaults to 1)
        anchor_node   : Node number corresponding to anchor (defaults to 1)
        fairlead_node : Node number corresponding to fairlead (vessel) node (defaults to 2)
        flags         : see MAP reference (defaults to empty string ' ')
        
        OUTPUTS  : none
        """
        # Add flag for taut lines
        if self.tlpFlag:
            flags += ' LINEAR SPRING'
            
        self.finput.append('---------------------- LINE PROPERTIES ---------------------------------------')
        self.finput.append('Line    LineType  UnstrLen  NodeAnch  NodeFair  Flags')
        self.finput.append('(-)      (-)       (m)       (-)       (-)       (-)')
        self.finput.append('%d   %s   %f   %d   %d   %s' %
                          (line_number, discrete_inputs['mooring_type'], inputs['mooring_line_length'], anchor_node, fairlead_node, flags) )

        
    def write_solver_options(self, inputs):
        """Writes SOLVER OPTIONS section of input.map file, 
        which includes repeating node/line arrangement in even angular spacing around structure.
        
        INPUTS:
        ----------
        inputs        : dictionary of input parameters
        
        OUTPUTS  : none
        """
        # Unpack variables
        n_connect     = max(1, int(inputs['number_of_mooring_connections']))
        
        self.finput.append('---------------------- SOLVER OPTIONS-----------------------------------------')
        self.finput.append('Option')
        self.finput.append('(-)')
        self.finput.append('help')
        self.finput.append(' integration_dt 0')
        self.finput.append(' kb_default 3.0e6')
        self.finput.append(' cb_default 3.0e5')
        self.finput.append(' wave_kinematics ')
        self.finput.append('inner_ftol 1e-5')
        self.finput.append('inner_gtol 1e-5')
        self.finput.append('inner_xtol 1e-5')
        self.finput.append('outer_tol 1e-3')
        self.finput.append(' pg_cooked 10000 1')
        self.finput.append(' outer_fd')
        self.finput.append(' outer_bd')
        self.finput.append(' outer_cd')
        self.finput.append(' inner_max_its 200')
        self.finput.append(' outer_max_its 600')
        # Repeat the details for the one mooring line multiple times
        angles = np.linspace(0, 360, n_connect+1)[1:-1]
        line = 'repeat'
        for degree in angles:
            line += (' %d' % degree)
        self.finput.append(line)
        self.finput.append(' krylov_accelerator 3')
        self.finput.append(' ref_position 0.0 0.0 0.0')

        
    def write_input_file(self, inputs, discrete_inputs):
        """Writes SOLVER OPTIONS section of input.map file, 
        which includes repeating node/line arrangement in even angular spacing around structure.
        
        INPUTS:
        ----------
        inputs        : dictionary of input parameters
        
        OUTPUTS  : none
        """
        # Unpack variables
        fairleadDepth = inputs['fairlead']
        R_fairlead    = inputs['fairlead_radius']
        R_anchor      = inputs['anchor_radius']
        n_connect     = int(inputs['number_of_mooring_connections'])
        n_lines       = int(inputs['mooring_lines_per_connection'])
        ntotal        = n_connect * n_lines
            
        
        # Open the map input file
        self.finput = []

        # Write the "Line Dictionary" section
        self.write_line_dictionary(inputs, discrete_inputs)

        # Write the "Node Properties" section
        self.write_node_properties_header()
        # One end on sea floor the other at fairlead
        self.write_node_properties(1, "VESSEL", R_fairlead, 0, -fairleadDepth)
        if n_lines > 1:
            angles  = np.linspace(0, 2*np.pi, ntotal+1)[:n_lines]
            angles -= np.mean(angles)
            anchorx = R_anchor * np.cos( angles )
            anchory = R_anchor * np.sin( angles )
            for k in range(n_lines):
                self.write_node_properties(k+2, "FIX", anchorx[k], anchory[k], None)
        else:
            self.write_node_properties(2, "FIX", R_anchor, 0, None)

        # Write the "Line Properties" section
        for k in range(n_lines):
            self.write_line_properties(inputs, discrete_inputs, line_number=k+1, anchor_node=k+2, fairlead_node=1)

        # Write the "Solve Options" section
        self.write_solver_options(inputs)
        
        
    def runMAP(self, inputs, discrete_inputs, outputs):
        """Writes MAP input file, executes, and then queries MAP to find 
        maximum loading and displacement from vessel displacement around all 360 degrees
        
        INPUTS:
        ----------
        inputs   : dictionary of input parameters
        outputs : dictionary of output parameters
        
        OUTPUTS  : none (multiple unknown dictionary values set)
        """
        # Unpack variables
        rhoWater      = inputs['water_density']
        waterDepth    = inputs['water_depth']
        fairleadDepth = inputs['fairlead']
        Dmooring      = inputs['mooring_diameter']
        offset        = inputs['max_offset']
        heel          = inputs['operational_heel']
        gamma         = inputs['gamma_f']
        n_connect     = int(inputs['number_of_mooring_connections'])
        n_lines       = int(inputs['mooring_lines_per_connection'])
        ntotal        = n_connect * n_lines

        # Write the mooring system input file for this design
        self.write_input_file(inputs, discrete_inputs)

        # Initiate MAP++ for this design
        mymap = pyMAP( )
        #mymap.ierr = 0
        mymap.map_set_sea_depth(waterDepth)
        mymap.map_set_gravity(gravity)
        mymap.map_set_sea_density(rhoWater)
        mymap.read_list_input(self.finput)
        mymap.init( )

        # Get the stiffness matrix at neutral position
        mymap.displace_vessel(0, 0, 0, 0, 0, 0)
        mymap.update_states(0.0, 0)
        K = mymap.linear(1e-4) # Input finite difference epsilon
        outputs['mooring_stiffness'] = np.array( K )
        mymap.displace_vessel(0, 0, 0, 0, 0, 0)
        mymap.update_states(0.0, 0)
        
        # Get the vertical load on the structure and plotting data
        F_neutral = np.zeros((NLINES_MAX, 3))
        plotMat   = np.zeros((NLINES_MAX, NPTS_PLOT, 3))
        nptsMOI   = 100
        xyzpts    = np.zeros((ntotal, nptsMOI, 3)) # For MOI calculation
        for k in range(ntotal):
            (F_neutral[k,0], F_neutral[k,1], F_neutral[k,2]) = mymap.get_fairlead_force_3d(k)
            plotMat[k,:,0] = mymap.plot_x(k, NPTS_PLOT)
            plotMat[k,:,1] = mymap.plot_y(k, NPTS_PLOT)
            plotMat[k,:,2] = mymap.plot_z(k, NPTS_PLOT)
            xyzpts[k,:,0]  = mymap.plot_x(k, nptsMOI)
            xyzpts[k,:,1]  = mymap.plot_y(k, nptsMOI)
            xyzpts[k,:,2]  = mymap.plot_z(k, nptsMOI)
            if self.tlpFlag:
                # Seems to be a bug in the plot arrays from MAP++ for plotting output with taut lines
                plotMat[k,:,2] = np.linspace(-fairleadDepth, -waterDepth, NPTS_PLOT)
                xyzpts[k,:,2]  = np.linspace(-fairleadDepth, -waterDepth, nptsMOI)
        outputs['mooring_neutral_load'] = F_neutral
        outputs['mooring_plot_matrix']  = plotMat
        
        # Fine line segment length, ds = sqrt(dx^2 + dy^2 + dz^2)
        xyzpts_dx = np.gradient(xyzpts[:,:,0], axis=1)
        xyzpts_dy = np.gradient(xyzpts[:,:,1], axis=1)
        xyzpts_dz = np.gradient(xyzpts[:,:,2], axis=1)
        xyzpts_ds = np.sqrt(xyzpts_dx**2 + xyzpts_dy**2 + xyzpts_dz**2)

        # Initialize inertia tensor integrands in https://en.wikipedia.org/wiki/Moment_of_inertia#Inertia_tensor
        # Taking MOI relative to body centerline at fairlead depth
        r0 = np.array([0.0, 0.0, -fairleadDepth])
        R  = np.zeros((ntotal, nptsMOI, 6))
        for ii in range(nptsMOI):
            for k in range(ntotal):
                r = xyzpts[k,ii,:] - r0
                R[k,ii,:] = unassembleI(np.dot(r,r)*np.eye(3) - np.outer(r,r))
        Imat = self.wet_mass_per_length * np.trapz(R, x=xyzpts_ds[:,:,np.newaxis], axis=1)
        outputs['mooring_moments_of_inertia'] = np.abs( Imat.sum(axis=0) )

        # Get the restoring moment at maximum angle of heel
        # Since we don't know the substucture CG, have to just get the forces of the lines now and do the cross product later
        # We also want to allow for arbitraty wind direction and yaw of rotor relative to mooring lines, so we will compare
        # pitch and roll forces as extremes
        # TODO: This still isgn't quite the same as clocking the mooring lines in different directions,
        # which is what we want to do, but that requires multiple input files and solutions
        Fh = np.zeros((NLINES_MAX,3))
        mymap.displace_vessel(0, 0, 0, 0, heel, 0)
        mymap.update_states(0.0, 0)
        for k in range(ntotal):
            Fh[k][0], Fh[k][1], Fh[k][2] = mymap.get_fairlead_force_3d(k)

        outputs['operational_heel_restoring_force'] = Fh

        # Get angles by which to find the weakest line
        dangle  = 2.0
        angles  = np.deg2rad( np.arange(0.0, 360.0, dangle) )
        nangles = len(angles)

        # Get restoring force at weakest line at maximum allowable offset
        # Will global minimum always be along mooring angle?
        max_tension   = 0.0
        max_angle     = None
        min_angle     = None
        F_max_tension = None
        F_min         = np.inf
        T = np.zeros((NLINES_MAX,))
        F = np.zeros((NLINES_MAX,))
        # Loop around all angles to find weakest point
        for a in angles:
            # Unit vector and offset in x-y components
            idir  = np.array([np.cos(a), np.sin(a)])
            surge = offset * idir[0]
            sway  = offset * idir[1]

            # Get restoring force of offset at this angle
            mymap.displace_vessel(surge, sway, 0, 0, 0, 0) # 0s for z, angles
            mymap.update_states(0.0, 0)
            for k in range(ntotal):
                # Force in x-y-z coordinates
                fx,fy,fz = mymap.get_fairlead_force_3d(k)
                T[k]     = np.sqrt(fx*fx + fy*fy + fz*fz)
                # Total restoring force
                F[k]     = np.dot([fx, fy], idir)
                
            # Check if this is the weakest direction (highest tension)
            tempMax = T.max()
            if tempMax > max_tension:
                max_tension   = tempMax
                F_max_tension = F.sum()
                max_angle     = a
            if F.sum() < F_min:
                F_min     = F.sum()
                min_angle = a
                
        # Store the weakest restoring force when the vessel is offset the maximum amount
        outputs['max_offset_restoring_force'] = F_min

        # Check for good convergence
        if (plotMat[0,-1,-1] + fairleadDepth) > 1.0:
            outputs['axial_unity'] = 1e30
        else:
            outputs['axial_unity'] = gamma * max_tension / self.min_break_load

        mymap.end()

        
    def compute_cost(self, inputs, discrete_inputs, outputs):
        """Computes cost, based on mass scaling, of mooring system.
        
        INPUTS:
        ----------
        inputs   : dictionary of input parameters
        outputs : dictionary of output parameters
        
        OUTPUTS  : none (mooring_cost/mass unknown dictionary values set)
        """
        # Unpack variables
        rhoWater      = inputs['water_density']
        L_mooring     = inputs['mooring_line_length']
        anchorType    = discrete_inputs['anchor_type']
        costFact      = inputs['mooring_cost_factor']
        n_connect     = int(inputs['number_of_mooring_connections'])
        n_lines       = int(inputs['mooring_lines_per_connection'])
        ntotal        = n_connect * n_lines
        
        # Cost of anchors
        if type(anchorType) == type(''): anchorType = Anchor[anchorType.upper()]
        if anchorType == Anchor['DRAGEMBEDMENT']:
            anchor_rate = 1e-3 * self.min_break_load / gravity / 20*2000
        elif anchorType  == Anchor['SUCTIONPILE']:
            anchor_rate = 150000.* np.sqrt(1e-3*self.min_break_load/gravity/1250.)
        else:
            raise ValueError('Anchor Type must be DRAGEMBEDMENT or SUCTIONPILE')
        anchor_total = anchor_rate*ntotal

        # Cost of all of the mooring lines
        legs_total = ntotal * self.cost_per_length * L_mooring

        # Total summations
        outputs['anchor_cost']  = anchor_total
        outputs['mooring_cost'] = costFact*(legs_total + anchor_total)
        outputs['mooring_mass'] = self.wet_mass_per_length*L_mooring*ntotal
        outputs['number_of_mooring_lines'] = ntotal




# http://stackoverflow.com/a/505457 with GB modifications
class Enum(object):
    def __init__(self, names, separator=None):
        if isinstance(names, type('')):
            names = names.upper().split(separator)
            self.data  = tuple(enumerate(names))
        elif isinstance(names, list):
            for i in names:
                assert len(i) == 2, 'Incorrect input data'
                assert isinstance(i[0], int), 'First index must be an integer'
                assert isinstance(i[1], type('')) or isinstance(i[1], type(u'')), 'Second index must be a string'
            self.data = names
        for i in self.data:
            setattr(self, i[1], i[0]) # name, value
        #for value, name in enumerate(names):
        #    setattr(self, name, value)
    def tuples(self):
        return self.data
    def __getitem__(self, i):
        if isinstance(i, type('')) or isinstance(i, type(u'')):
            return getattr(self, i)
        else:
            return self.data[i][1]
    def __len__(self):
        return len(self.data)



# http://stackoverflow.com/a/1695250
"""
#def enum(*sequential, **named):
#    enums = dict(zip(sequential, range(len(sequential))), **named)
#    return type('Enum', (), enums)
class Enum:
    def __init__(self, names):
        self.names = names.split()
        make(self.names)

    def make(*sequential, **named):
        enums = dict(zip(sequential, range(len(sequential))), **named)
        return type('Enum', (), enums)
"""


# Taken from http://norvig.com/python-iaq.html
"""
class Enum:

    # Create an enumerated type, then add var/value pairs to it.
    # The constructor and the method .ints(names) take a list of variable names,
    # and assign them consecutive integers as values.    The method .strs(names)
    # assigns each variable name to itself (that is variable 'v' has value 'v').
    # The method .vals(a=99, b=200) allows you to assign any value to variables.
    # A 'list of variable names' can also be a string, which will be .split().
    # The method .end() returns one more than the maximum int value.
    # Example: opcodes = Enum("add sub load store").vals(illegal=255).
  
    def __init__(self, names=[]): self.ints(names)

    def set(self, var, val):
        # Set var to the value val in the enum.
        if var in vars(self).keys(): raise AttributeError("duplicate var in enum")
        if val in vars(self).values(): raise ValueError("duplicate value in enum")
        vars(self)[var] = val
        return self
  
    def strs(self, names):
        # Set each of the names to itself (as a string) in the enum.
        for var in self._parse(names): self.set(var, var)
        return self

    def ints(self, names):
        # Set each of the names to the next highest int in the enum.
        for var in self._parse(names): self.set(var, self.end())
        return self

    def vals(self, **entries):
        # Set each of var=val pairs in the enum.
        for (var, val) in entries.items(): self.set(var, val)
        return self

    def end(self):
        # One more than the largest int value in the enum, or 0 if none.
        try: return max([x for x in vars(self).values() if type(x)==type(0)]) + 1
        except ValueError: return 0

    def _parse(self, names):
        # If names is a string, parse it as a list of names.
        if type(names) == type(""): return names.split()
        else: return names
"""
